# Alternative to the bash script to run the main openephys job
import os
import shutil
from pathlib import Path

from aind_data_transfer.jobs.openephys_job import run_job

# ADD TUPLES OF PATH STRINGS TO RAW DATA DIRECTORY AND BEHAVIOR DIRECTORY
# SET BEHAVIOR DIR TO NONE IF INCLUDED IN RAW DATA DIRECTORY (LEGACY)
LIST_OF_DIRECTORIES_TO_PROCESS = [
    (
        f"${str(Path(os.getcwd()))}/tests/resources/"
        f"v0.6.x_neuropixels_multiexp_multistream",
        None,
    )
]

# DELETE TEMP COMPRESSED DATA FOLDER AFTER UPLOAD? SET TO FALSE FOR TESTING.
DELETE_TEMP_COMPRESSED_DATA_FOLDER = True

# Path to the default configs. Change it to point to your config file.
# The contents of config file are not expected to change too often.
CONFIG_LOCATION = str(
    Path(os.getcwd())
    / "tests"
    / "resources"
    / "test_configs"
    / "ephys_upload_job_test_configs.yml"
)


if __name__ == "__main__":
    for list_item in LIST_OF_DIRECTORIES_TO_PROCESS:
        raw_data_source = list_item[0]
        behavior_dir = list_item[1]
        if behavior_dir is not None:
            dest_data_dir = run_job(
                [
                    "-c",
                    CONFIG_LOCATION,
                    "-r",
                    raw_data_source,
                    "-b",
                    behavior_dir,
                ]
            )
        else:
            dest_data_dir = run_job(
                ["-c", CONFIG_LOCATION, "-r", raw_data_source]
            )
        if DELETE_TEMP_COMPRESSED_DATA_FOLDER:
            shutil.rmtree(dest_data_dir)



"""Module for ephys data readers."""
from enum import Enum

import numpy as np
from spikeinterface import extractors as se


class DataReader(Enum):
    """Enum for readers."""

    OPENEPHYS = "openephys"


class EphysReaders:
    """This class contains the methods to retrieve a reader."""

    readers = [member.value for member in DataReader]

    class RecordingBlockPrefixes(Enum):
        """Enum for types of recording blocks."""

        # TODO: Convert these to regex patterns?
        neuropix = "Neuropix"
        nidaq = "NI-DAQ"

    class SourceRegexPatterns(Enum):
        """Enum for regex patterns the source folder name should match"""

        subject_datetime = r"(\d+)_(\d{4}-\d{2}-\d{2})_(\d{2}-\d{2}-\d{2})"
        ecephys_subject_datetime = (
            r"ecephys_(\d+)_(\d{4}-\d{2}-\d{2})_(\d{2}-\d{2}-\d{2})"
        )

    @staticmethod
    def get_read_blocks(reader_name, input_dir):
        """
        Given a reader name and input directory, generate a stream of
        recording blocks.
        Args:
            reader_name (str): Matches one of the names Readers enum
            input_dir (Path): Directory of ephys data to read.
        Returns:
            A generator of read blocks. A read_block is dict of
            {'recording', 'experiment_name', 'stream_name'}.

        """
        if reader_name == DataReader.OPENEPHYS.value:
            nblocks = se.get_neo_num_blocks(reader_name, input_dir)
            stream_names, stream_ids = se.get_neo_streams(
                reader_name, input_dir
            )
            # load first stream to map block_indices to experiment_names
            rec_test = se.read_openephys(
                input_dir, block_index=0, stream_name=stream_names[0]
            )
            record_node = list(rec_test.neo_reader.folder_structure.keys())[0]
            experiments = rec_test.neo_reader.folder_structure[record_node][
                "experiments"
            ]
            exp_ids = list(experiments.keys())
            experiment_names = [
                experiments[exp_id]["name"] for exp_id in sorted(exp_ids)
            ]
            for block_index in range(nblocks):
                for stream_name in stream_names:
                    rec = se.read_openephys(
                        input_dir,
                        stream_name=stream_name,
                        block_index=block_index,
                    )
                    yield (
                        {
                            "recording": rec,
                            "experiment_name": experiment_names[block_index],
                            "stream_name": stream_name,
                        }
                    )
        else:
            raise Exception(
                f"Unknown reader: {reader_name}. "
                f"Please select one of {EphysReaders.readers}"
            )

    @staticmethod
    def get_streams_to_clip(reader_name, input_dir):
        stream_names, stream_ids = se.get_neo_streams(reader_name, input_dir)
        for dat_file in input_dir.glob("**/*.dat"):
            oe_stream_name = dat_file.parent.name
            si_stream_name = [
                stream_name
                for stream_name in stream_names
                if oe_stream_name in stream_name
            ][0]
            n_chan = se.read_openephys(
                input_dir, block_index=0, stream_name=si_stream_name
            ).get_num_channels()
            data = np.memmap(
                dat_file, dtype="int16", order="C", mode="r"
            ).reshape(-1, n_chan)
            yield {
                "data": data,
                "relative_path_name": str(dat_file.relative_to(input_dir)),
                "n_chan": n_chan,
            }



import os
import shutil
import tempfile
import unittest
from pathlib import Path

import dask.array
import h5py
import numpy as np
from tifffile import tifffile

from aind_data_transfer.util.io_utils import (
    DataReaderFactory,
    ImarisReader,
    TiffReader,
)

# TODO: make test fixtures instead of constants?
IM_SHAPE = (64, 128, 128)
IM_DTYPE = np.dtype("uint16")


def _write_test_tiffs(folder, n=1):
    paths = []
    for i in range(n):
        a = np.ones(IM_SHAPE, dtype=IM_DTYPE)
        path = os.path.join(folder, f"data_{i}.tif")
        paths.append(path)
        tifffile.imwrite(path, a, imagej=True)
    return paths


def _write_test_h5(folder, n=1):
    paths = []
    for i in range(n):
        a = np.ones(IM_SHAPE, dtype=IM_DTYPE)
        path = os.path.join(folder, f"data_{i}.h5")
        paths.append(path)
        with h5py.File(path, "w") as f:
            f.create_dataset(ImarisReader.DEFAULT_DATA_PATH, data=a, chunks=True)
            # Write origin metadata
            dataset_info = f.create_group("DataSetInfo/Image")
            dataset_info.attrs["X"] = np.array(
                list(str(IM_SHAPE[2])), dtype="S"
            )
            dataset_info.attrs["Y"] = np.array(
                list(str(IM_SHAPE[1])), dtype="S"
            )
            dataset_info.attrs["Z"] = np.array(
                list(str(IM_SHAPE[0])), dtype="S"
            )
    return paths


class TestTiffReader(unittest.TestCase):
    def setUp(self):
        self._image_dir = Path(__file__).parent / "resources/imaging/data/tiff"
        self._image_dir.mkdir(parents=True, exist_ok=True)
        self._image_path = _write_test_tiffs(self._image_dir, n=1)[0]
        self._reader = TiffReader(self._image_path)

    def tearDown(self) -> None:
        self._reader.close()
        shutil.rmtree(self._image_dir)

    def test_get_filepath(self):
        self.assertEqual(self._image_path, self._reader.get_filepath())

    def test_as_array(self):
        a = self._reader.as_array()
        self.assertIsInstance(a, np.ndarray)
        self.assertEqual(IM_SHAPE, a.shape)
        self.assertTrue(np.all(a == 1))
        self.assertEqual(IM_DTYPE, a.dtype)

    def test_as_dask_array(self):
        d = self._reader.as_dask_array()
        self.assertIsInstance(d, dask.array.Array)
        self.assertEqual(IM_SHAPE, d.shape)
        self.assertTrue(np.all(d == 1))
        self.assertEqual(IM_DTYPE, d.dtype)

    def test_get_shape(self):
        self.assertEqual(IM_SHAPE, self._reader.get_shape())

    def test_get_chunks(self):
        chunks = self._reader.get_chunks()
        self.assertEqual(3, len(chunks))
        self.assertTrue(all(isinstance(c, int) for c in chunks))
        self.assertTrue(all(c >= 1 for c in chunks))

    def test_get_dtype(self):
        self.assertEqual(IM_DTYPE, self._reader.get_dtype())

    def test_get_itemsize(self):
        self.assertEqual(IM_DTYPE.itemsize, self._reader.get_itemsize())


class TestHDF5Reader(unittest.TestCase):
    def setUp(self):
        self._temp_dir = tempfile.TemporaryDirectory()
        self._image_dir = Path(self._temp_dir.name) / "images"
        os.makedirs(self._image_dir, exist_ok=True)
        self._image_path = _write_test_h5(self._image_dir, n=1)[0]
        self._reader = ImarisReader(self._image_path)

    def tearDown(self) -> None:
        self._reader.close()
        self._temp_dir.cleanup()

    def test_constructor(self):
        self.assertEqual(self._reader.filepath, self._image_path)
        self.assertIsNotNone(self._reader.get_handle())

    def test_get_filepath(self):
        self.assertEqual(self._image_path, self._reader.get_filepath())

    def test_as_array(self):
        a = self._reader.as_array()
        self.assertIsInstance(a, np.ndarray)
        self.assertEqual(IM_SHAPE, a.shape)
        self.assertTrue(np.all(a == 1))
        self.assertEqual(IM_DTYPE, a.dtype)

    def test_as_dask_array(self):
        d = self._reader.as_dask_array(chunks=(32, 32, 32))
        self.assertIsInstance(d, dask.array.Array)
        self.assertEqual(IM_SHAPE, d.shape)
        self.assertTrue(np.all(d == 1))
        self.assertEqual(IM_DTYPE, d.dtype)

    def test_get_shape(self):
        self.assertEqual(IM_SHAPE, self._reader.get_shape())

    def test_get_chunks(self):
        chunks = self._reader.get_chunks()
        self.assertEqual(3, len(chunks))
        self.assertTrue(all(isinstance(c, int) for c in chunks))
        self.assertTrue(all(c >= 1 for c in chunks))

    def test_get_itemsize(self):
        self.assertEqual(IM_DTYPE.itemsize, self._reader.get_itemsize())

    def test_get_handle(self):
        self.assertIsNotNone(self._reader.get_handle())

    def test_close(self):
        self._reader.close()
        self.assertIsNone(self._reader.handle)
        self.assertRaises(TypeError, self._reader.as_array)


class TestDataReaderFactor(unittest.TestCase):
    def setUp(self):
        self._temp_dir = tempfile.TemporaryDirectory()
        self._image_dir = Path(self._temp_dir.name) / "images"
        os.makedirs(self._image_dir, exist_ok=True)
        self._h5_path = _write_test_h5(self._image_dir, n=1)[0]
        self._tiff_path = _write_test_tiffs(self._image_dir, n=1)[0]

    def tearDown(self) -> None:
        self._temp_dir.cleanup()

    def test_create_hdf5reader(self):
        reader = DataReaderFactory().create(self._h5_path)
        self.assertIsInstance(reader, ImarisReader)
        reader.close()

    def test_create_tiffreader(self):
        reader = DataReaderFactory().create(self._tiff_path)
        self.assertIsInstance(reader, TiffReader)
        reader.close()


if __name__ == "__main__":
    unittest.main()



"""Tests ephys configurations are set properly"""
import os
import unittest
from pathlib import Path
from unittest import mock

from numcodecs import Blosc

from aind_data_transfer.config_loader.ephys_configuration_loader import (
    EphysJobConfigurationLoader,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"


class TestEphysJobConfigs(unittest.TestCase):
    """Tests ephys job pipeline methods"""

    @mock.patch(
        "aind_data_transfer.config_loader.ephys_configuration_loader."
        "EphysJobConfigurationLoader._get_endpoints"
    )
    def test_conf_loads(self, mocked_get_endpoints):
        """Basic config loads test"""

        mocked_get_endpoints.return_value = {"codeocean_trigger_capsule": None}

        raw_data_dir = (
            "tests/resources/v0.6.x_neuropixels_multiexp_multistream"
        )
        dest_data_dir = (
            "tests/resources/new/v0.6.x_neuropixels_multiexp_multistream"
        )

        expected_configs = {
            "jobs": {
                "clip": True,
                "compress": True,
                "attach_metadata": False,
                "upload_to_s3": True,
                "upload_to_gcp": False,
                "trigger_codeocean_job": False,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": dest_data_dir,
                "s3_bucket": "some-s3-bucket",
                "s3_prefix": "v0.6.x_neuropixels_multiexp_multistream",
                "gcp_bucket": "some-gcp-bucket",
                "gcp_prefix": "test_20221001",
                "codeocean_domain": "https://acmecorp.codeocean.com",
                "code_repo_location": "https://location_of_code_repo",
                "metadata_service_url": "http://some-url",
            },
            "aws_secret_names": {
                "code_ocean_api_token_name": "secret_name_for_api_token",
                "region": "us-west-2",
                "video_encryption_password": "video_encryption_password",
            },
            "data": {"name": "openephys"},
            "clip_data_job": {
                "clip_kwargs": {},
            },
            "compress_data_job": {
                "write_kwargs": {
                    "n_jobs": -1,
                    "chunk_duration": "1s",
                    "progress_bar": True,
                },
                "format_kwargs": {},
                "compressor": {
                    "compressor_name": "blosc",
                    "kwargs": {"shuffle": Blosc.BITSHUFFLE},
                },
                "scale_params": {"chunk_size": 20},
            },
            "upload_data_job": {"dryrun": True},
            "trigger_codeocean_job": {
                "job_type": "openephys",
                "bucket": "some-s3-bucket",
                "prefix": "v0.6.x_neuropixels_multiexp_multistream",
            },
            "logging": {"level": "INFO"},
        }
        conf_file_path = CONFIGS_DIR / "ephys_upload_job_test_configs.yml"

        args = ["-c", str(conf_file_path)]

        loaded_configs = EphysJobConfigurationLoader().load_configs(args)
        self.assertEqual(loaded_configs, expected_configs)

    @mock.patch(
        "aind_data_transfer.config_loader.ephys_configuration_loader."
        "EphysJobConfigurationLoader._get_endpoints"
    )
    def test_endpoints_are_resolved(self, mocked_get_endpoints):
        """Tests default endpoints are resolved correctly"""

        mocked_get_endpoints.return_value = {"codeocean_trigger_capsule": None}

        raw_data_dir = "/some/random/folder/625463_2022-10-06_10-14-25"
        expected_configs = {
            "jobs": {
                "clip": True,
                "compress": True,
                "attach_metadata": False,
                "upload_to_s3": True,
                "upload_to_gcp": True,
                "trigger_codeocean_job": False,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": "ecephys_625463_2022-10-06_10-14-25",
                "s3_bucket": "some-s3-bucket",
                "s3_prefix": "ecephys_625463_2022-10-06_10-14-25",
                "gcp_bucket": "some-gcp-bucket",
                "gcp_prefix": "ecephys_625463_2022-10-06_10-14-25",
                "codeocean_domain": "https://acmecorp.codeocean.com",
                "code_repo_location": "https://location_of_code_repo",
                "metadata_service_url": "http://some-url",
            },
            "aws_secret_names": {
                "code_ocean_api_token_name": "secret_name_for_api_token",
                "region": "us-west-2",
                "video_encryption_password": "secret_name_for_vid_password",
            },
            "data": {"name": "openephys", "subject_id": "625463"},
            "clip_data_job": {"clip_kwargs": {}},
            "compress_data_job": {
                "write_kwargs": {
                    "n_jobs": -1,
                    "chunk_duration": "1s",
                    "progress_bar": True,
                },
                "format_kwargs": {},
                "compressor": {
                    "compressor_name": "wavpack",
                    "kwargs": {"level": 3},
                },
                "scale_params": {},
            },
            "upload_data_job": {"dryrun": True},
            "trigger_codeocean_job": {
                "job_type": "openephys",
                "bucket": "some-s3-bucket",
                "prefix": "ecephys_625463_2022-10-06_10-14-25",
            },
            "logging": {"level": "INFO"},
        }

        conf_file_path1 = CONFIGS_DIR / "example_configs_src_pattern1.yml"
        conf_file_path2 = CONFIGS_DIR / "example_configs_src_pattern2.yml"
        args1 = ["-c", str(conf_file_path1)]
        args2 = ["-c", str(conf_file_path2)]
        loaded_configs1 = EphysJobConfigurationLoader().load_configs(args1)
        loaded_configs2 = EphysJobConfigurationLoader().load_configs(args2)
        self.assertEqual(loaded_configs1, expected_configs)
        self.assertEqual(loaded_configs2, expected_configs)


if __name__ == "__main__":
    unittest.main()



"""Module for basic compression classes."""
import os
from enum import Enum
from pathlib import Path
from typing import List, Optional

import pyminizip
from tqdm import tqdm


class ZipCompressor:
    """Class to handle basic zip compression."""

    def __init__(
        self,
        compression_level: int = 5,
        encryption_key: Optional[str] = None,
        display_progress_bar: bool = False,
    ) -> None:
        """
        Creates a video compressor with compression level and encryption key
        Parameters
        ----------
        compression_level : int
          Integer between 1 and 9. Default is 5.
        encryption_key : Optional[str]
          Optional password to use. Default is None.
        display_progress_bar : bool
          Display the progress bar for the compression. Default is False.
        """
        self.compression_level = compression_level
        self.encryption_key = encryption_key
        self.display_progress_bar = display_progress_bar

    def compress_dir(
        self,
        input_dir: Path,
        output_dir: Path,
        skip_dirs: Optional[List[Path]] = None,
    ):
        """
        Compress the contents of the input folder and save it as a zipped
        folder in an output directory. Can optionally provide a list of
        subdirectories and files to ignore (no wildcard support yet).
        Parameters
        ----------
        input_dir : Path
          The location of the folder to compress
        output_dir : Path
          The location where to save the zipped folder
        skip_dirs : Optional[List[Path]]
          An optional list of directories or files that will be ignored

        Returns
        -------
        None

        """

        # TODO: There's probably a more efficient way than this
        def skip_path(path_to_check, list_of_paths) -> bool:
            """Utility method to check whether to skip a path"""
            if list_of_paths is None:
                return False
            else:
                for path in list_of_paths:
                    if str(path_to_check).startswith(str(path)):
                        return True
                return False

        file_names = []
        file_prefixes = []
        for root, dirs, files in os.walk(input_dir):
            for file in files:
                raw_file_path = os.path.join(root, file)
                if not skip_path(raw_file_path, list_of_paths=skip_dirs):
                    prefix = str(root).replace(str(input_dir.parent), "", 1)[
                        1:
                    ]
                    file_names.append(raw_file_path)
                    file_prefixes.append(prefix)
        total_file_count = len(file_names)
        pbar = tqdm(
            total=total_file_count, disable=(not self.display_progress_bar)
        )
        pyminizip.compress_multiple(
            file_names,
            file_prefixes,
            str(output_dir),
            self.encryption_key,
            self.compression_level,
            lambda x: pbar.update(x),
        )
        pbar.close()
        return None


class VideoCompressor:
    """Class to handle video compression and encryption."""

    def __init__(
        self, compression_level: int = 5, encryption_key: Optional[str] = None
    ) -> None:
        """
        Creates a video compressor with compression level and encryption key
        Parameters
        ----------
        compression_level : int
          Integer between 1 and 9. Default is 5.
        encryption_key : Optional[str]
          Optional password to use. Default is None.
        """
        self.compression_level = compression_level
        self.encryption_key = encryption_key

    class VideoFileTypes(Enum):
        """Enum for types of video to compress"""

        MPEG_4 = ".mp4"
        QUICKTIME_MOVIE = ".mov"
        WINDOWSMEDIA_VIEWER = ".wmv"
        AUDIO_VIDEO_INTERLEAVE = ".avi"

    video_file_extensions = tuple([member.value for member in VideoFileTypes])

    def compress_all_videos_in_dir(self, video_dir: Path) -> None:
        """
        Compress and optionally encrypt video files in a directory
        Parameters
        ----------
        video_dir : Path
          Directory where video files are stored.

        Returns
        -------
        None

        """

        for root, dirs, files in os.walk(video_dir):
            for file in files:
                if file.endswith(self.video_file_extensions):
                    raw_file_path = os.path.join(root, file)
                    zip_file_path = os.path.join(root, file) + ".zip"
                    pyminizip.compress(
                        str(raw_file_path),
                        None,
                        str(zip_file_path),
                        self.encryption_key,
                        self.compression_level,
                    )
                    os.remove(raw_file_path)



"""This module contains the api to retrieve a reader for imaging data.
"""
import os
import re
from enum import Enum
from pathlib import Path
from typing import List, Union

PathLike = Union[str, Path]


class ImagingReaders:
    """This class contains the methods to retrieve a reader for aind imaging data."""

    class Readers(Enum):
        """Enum for readers."""

        dispim = "diSPIM"
        exaspim = "exaSPIM"
        mesospim = "mesoSPIM"
        generic = "micr"

    readers = [member.value for member in Readers]

    class SourceRegexPatterns(Enum):
        """Enum for regex patterns the source folder name should match"""

        exaspim_acquisition = r"exaSPIM_([-A-Z0-9.]+)_(\d{4})-(\d{2})-(\d{2})_(\d{2})-(\d{2})-(\d{2})"
        mesospim_acquisition = r"mesoSPIM_([-A-Z0-9.]+)_(\d{4})-(\d{2})-(\d{2})_(\d{2})-(\d{2})-(\d{2})"
        dispim_acquisition = r"diSPIM_([-A-Z0-9.]+)_(\d{4})-(\d{2})-(\d{2})_(\d{2})-(\d{2})-(\d{2})"

    @staticmethod
    def get_raw_data_dir(reader_name, input_dir):
        if reader_name not in ImagingReaders.readers:
            raise Exception(
                f"Unknown readers: {reader_name}. "
                f"Please select one of {ImagingReaders.readers}"
            )
        raw_data_dir = Path(input_dir) / reader_name
        if not raw_data_dir.is_dir():
            raise FileNotFoundError(
                f"Raw data directory not found: {raw_data_dir}"
            )
        return raw_data_dir

    @staticmethod
    def get_reader_name(input_dir):
        # re.search does not work with Path objects
        input_dir = str(input_dir)
        if (
            re.search(ImagingReaders.Readers.exaspim.value, input_dir)
            is not None
        ):
            return ImagingReaders.Readers.exaspim.value
        elif (
            re.search(ImagingReaders.Readers.mesospim.value, input_dir)
            is not None
        ):
            return ImagingReaders.Readers.mesospim.value
        elif (
            re.search(ImagingReaders.Readers.dispim.value, input_dir)
            is not None
        ):
            return ImagingReaders.Readers.dispim.value
        else:
            return ImagingReaders.Readers.generic.value


class SmartSPIMReader:
    """Reader for smartspim datasets"""

    class RegexPatterns(Enum):
        """Enum for regex patterns for the smartSPIM data"""

        # regex expressions for not structured smartspim datasets
        capture_date_regex = r"(20[0-9]{2}([0-9][0-9]{1})([0-9][0-9]{1}))"
        capture_time_regex = r"(_(\d{2})_(\d{2})_(\d{2})_)"
        capture_mouse_id = r"(_(\d+|[a-zA-Z]*\d+)$)"

        # Regular expression for smartspim datasets
        smartspim_regex = r"SmartSPIM_(\d+|[a-zA-Z]*\d+)_(20\d{2}-(\d\d{1})-(\d\d{1}))_((\d{2})-(\d{2})-(\d{2}))"
        raw_smartspim_regex = r"((20[0-9]{2}[0-9][0-9]{1}[0-9][0-9]{1})_(\d{2}_\d{2}_\d{2})_(\d+|[a-zA-Z]*\d+)$)"

        # Regex expressions for inner folders inside root
        regex_channels = r"Ex_(\d{3})_Em_(\d{3})$"
        regex_channels_MIP = r"Ex_(\d{3})_Em_(\d{3}_MIP)$"
        regex_files = r'[^"]*.(txt|ini|json)$'

    @staticmethod
    def read_raw_smartspim_folders(path: PathLike) -> List[str]:
        """
        Reads raw smartspim datasets in a folder

        Parameters
        -----------------
        path: PathLike
            Path where the datasets are located

        Returns
        -----------------
        List[str]
            List with the found raw smartspim datasets
        """
        smartspim_datasets = []

        if os.path.isdir(path):
            datasets = os.listdir(path)

            for dataset in datasets:
                if re.match(
                    SmartSPIMReader.RegexPatterns.raw_smartspim_regex.value,
                    dataset,
                ):
                    smartspim_datasets.append(dataset)

        else:
            raise ValueError(f"Path {path} is not a folder.")

        return smartspim_datasets

    @staticmethod
    def read_smartspim_folders(path: PathLike) -> List[str]:
        """
        Reads smartspim datasets in a folder
        based on data conventions

        Parameters
        -----------------
        path: PathLike
            Path where the datasets are located

        Returns
        -----------------
        List[str]
            List with the found smartspim datasets
        """
        smartspim_datasets = []

        if os.path.isdir(path):
            datasets = os.listdir(path)

            for dataset in datasets:
                if re.match(
                    SmartSPIMReader.RegexPatterns.smartspim_regex.value,
                    dataset,
                ):
                    smartspim_datasets.append(dataset)

        else:
            raise ValueError(f"Path {path} is not a folder.")

        return smartspim_datasets



import argparse
import logging

import google.cloud.exceptions

# Importing this alone doesn't work on HPC
# Must manually override HDF5_PLUGIN_PATH environment variable
# in each Dask worker
import hdf5plugin
import pandas as pd
from botocore.session import get_session
from numcodecs import blosc

from aind_data_transfer.gcs import create_client
from aind_data_transfer.transformations.ome_zarr import write_files
from aind_data_transfer.util.dask_utils import get_client
from aind_data_transfer.util.env_utils import find_hdf5plugin_path
from aind_data_transfer.util.file_utils import *

blosc.use_threads = False

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def get_blosc_codec(cname, clevel):
    opts = {
        "compressor": blosc.Blosc(
            cname=cname, clevel=clevel, shuffle=blosc.SHUFFLE
        ),
        "params": {
            "shuffle": blosc.SHUFFLE,
            "level": clevel,
            "name": f"blosc-{cname}",
        },
    }
    return opts


def output_valid(output: Union[str, os.PathLike]) -> bool:
    """
    Check if the output path is valid. If the path is a cloud storage location,
    check if the bucket exists and is accessible (but not necessarily writable) by the current user.

    Args:
        output: the output path or url

    Returns:
        True if the bucket exists and is accessible
    """
    if output.startswith("gs://"):
        _, bucket_name, __ = parse_cloud_url(output)
        client = create_client()
        try:
            _ = client.get_bucket(bucket_name)
            return True
        except google.cloud.exceptions.NotFound:
            LOGGER.error(f"Bucket not found: {bucket_name}")
            return False
        except Exception as e:
            LOGGER.error(f"Error retrieving bucket {bucket_name}: {e}")
            return False
    elif output.startswith("s3://"):
        _, bucket_name, __ = parse_cloud_url(output)
        session = get_session()
        client = session.create_client("s3")
        try:
            response = client.head_bucket(Bucket=bucket_name)
        except Exception as e:
            LOGGER.error(f"Error connecting to bucket {bucket_name}: {e}")
            return False
        status = response["ResponseMetadata"]["HTTPStatusCode"]
        if status == 200:
            return True
        else:
            LOGGER.error(
                f"Error accessing bucket: {bucket_name}"
                f"\nHTTP status code: {status}"
            )
            return False
    else:
        os.makedirs(output, exist_ok=True)
        return True


def ensure_metrics_file(metrics_file):
    if not metrics_file.endswith(".csv"):
        raise ValueError("metrics_file must be .csv")
    metrics_dir = os.path.dirname(os.path.abspath(metrics_file))
    if not os.path.isdir(metrics_dir):
        os.makedirs(metrics_dir, exist_ok=True)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input",
        type=str,
        help="directory of images to transcode to OME-Zarr. Each image"
        " is written to a separate group in the top-level zarr folder.",
    )
    parser.add_argument(
        "--output",
        type=str,
        help="output location of the OME-Zarr dataset, either a path or url (e.g., s3://my-bucket/my-zarr)",
    )
    parser.add_argument("--codec", type=str, default="zstd")
    parser.add_argument("--clevel", type=int, default=1)
    parser.add_argument(
        "--chunk_size", type=float, default=64, help="chunk size in MB"
    )
    parser.add_argument(
        "--chunk_shape",
        type=int,
        nargs="+",
        default=None,
        help="5D sequence of chunk dimensions, in TCZYX order",
    )
    parser.add_argument(
        "--n_levels", type=int, default=1, help="number of resolution levels"
    )
    parser.add_argument(
        "--scale_factor",
        type=float,
        default=2.0,
        help="scale factor for downsampling",
    )
    parser.add_argument(
        "--deployment",
        type=str,
        default="local",
        help="cluster deployment type",
    )
    parser.add_argument("--log_level", type=int, default=logging.INFO)
    parser.add_argument(
        "--metrics_file",
        type=str,
        default="tile-metrics.csv",
        help="output tile metrics csv file",
    )
    parser.add_argument(
        "--resume",
        default=False,
        action="store_true",
        help="resume processing",
    )
    parser.add_argument(
        "--exclude",
        default=[],
        type=str,
        nargs="+",
        help='filename patterns to exclude from transcoding, e.g., "*.tif", "*.memento", etc',
    )
    parser.add_argument(
        "--voxsize",
        type=str,
        default=None,
        help='Voxel size of the dataset as a string of floats in XYZ order, e.g. "0.3,0.3,1.0"',
    )
    parser.add_argument(
        "--bkg_img_dir",
        type=str,
        default=None,
        help="path to the background image folder. If given, background subtraction will be done for each converted image."
    )
    args = parser.parse_args()
    return args


def parse_voxel_size(voxsize_str):
    vsstr = voxsize_str.split(",")
    vs = []
    if len(vsstr) > 0:
        vs.append(float(vsstr[2]))
    else:
        vs.append(1.0)
    if len(vsstr) > 1:
        vs.append(float(vsstr[1]))
    else:
        vs.append(1.0)
    if len(vsstr) > 2:
        vs.append(float(vsstr[0]))
    else:
        vs.append(1.0)
    return vs


def main():
    args = parse_args()

    LOGGER.setLevel(args.log_level)

    worker_options = {
        "env": {
            "HDF5_PLUGIN_PATH": find_hdf5plugin_path(),
            "HDF5_USE_FILE_LOCKING": "FALSE",
            "OMP_NUM_THREADS": "1",
            "MALLOC_TRIM_THRESHOLD_": "0",
        }
    }
    client, _ = get_client(args.deployment, worker_options=worker_options)

    if not output_valid(args.output):
        LOGGER.error("Output path not valid, aborting.")
        return

    ensure_metrics_file(args.metrics_file)

    image_dir = Path(args.input)

    images = set(get_images(image_dir, exclude=args.exclude))
    LOGGER.info(f"Found {len(images)} images in {image_dir}")

    if not images:
        LOGGER.info(f"No images found, exiting.")
        return

    LOGGER.info(f"Writing {len(images)} images to OME-Zarr")
    LOGGER.info(f"Writing OME-Zarr to {args.output}")

    # If voxsize is None, we will
    # attempt to parse it from the image metadata
    voxsize = None
    if args.voxsize is not None:
        voxsize = parse_voxel_size(args.voxsize)

    overwrite = not args.resume

    opts = get_blosc_codec(args.codec, args.clevel)

    all_metrics = write_files(
        images,
        args.output,
        args.n_levels,
        args.scale_factor,
        overwrite,
        args.chunk_size,
        args.chunk_shape,
        voxsize,
        opts,
        bkg_img_dir=args.bkg_img_dir
    )

    df = pd.DataFrame.from_records(all_metrics)
    df.to_csv(args.metrics_file, index_label="test_number")

    client.shutdown()
    client.close()


if __name__ == "__main__":
    main()



import re
from enum import Enum
from typing import Union, List

import dask.array as da
import numpy as np


class Deinterleave:

    @staticmethod
    def deinterleave(
        a: Union[da.Array, np.ndarray],
        num_channels: int,
        axis: int,
    ) -> List[Union[da.Array, np.ndarray]]:
        """
        Deinterleave the channels of a dask or numpy array.

        Parameters
        ----------
        a : Union[da.Array, np.ndarray]
          The dask or numpy array.
        num_channels : int
          The number of arrays to extract
        axis: int
          The axis to deinterleave

        Returns
        -------
        List of dask or numpy arrays representing each deinterleaved channel
        """
        channels = []
        for offset in range(num_channels):
            s = [slice(None)] * a.ndim
            s[axis] = slice(offset, None, num_channels)
            channels.append(a[tuple(s)])
        return channels


class ChannelParser:

    class RegexPatterns(Enum):
        channel_pattern = r"ch_([0-9_]{3,})(_cam(0|1))?\."
        tile_xyz_pattern = r"([a-zA-Z0-9_]+)_[xX]_\d{4}_[yY]_\d{4}_[zZ]_\d{4}"

    @staticmethod
    def parse_channel_names(filepath: str):
        """
        Parse the channel wavelengths from a filepath

        Parameters
        ----------
        filepath: str
          the path to the interleaved image

        Returns
        -------
        List of channel wavelength strings, e.g., ["488", "561"]
        """
        filepath = str(filepath)
        m = re.search(ChannelParser.RegexPatterns.channel_pattern.value, filepath)
        if m is None:
            raise ValueError(f"file name does not match channel pattern: {filepath}")
        wavelengths = m.group(1).strip().split("_")
        return wavelengths

    @staticmethod
    def parse_tile_xyz_loc(filepath: str):
        """
        Parse the tile XYZ prefix from the filepath

        Parameters
        ----------
        filepath: str
          the path to the interleaved image

        Returns
        -------
        the tile coordinate prefix, e.g., "tile_X_0000_Y_0001_Z_0002"
        """
        filepath = str(filepath)
        m = re.search(ChannelParser.RegexPatterns.tile_xyz_pattern.value, filepath)
        if m is None:
            raise ValueError(f"file name does not match tile pattern: {filepath}")
        return m.group(0)


if __name__ == "__main__":
    path = "/home/cameron.arshadi/data/tile_X_0002_Y_0000_Z_0000_ch_488_561_688_cam1.tiff"
    names = ChannelParser.parse_channel_names(path)
    xyz_loc = ChannelParser.parse_tile_xyz_loc(path)
    print(names)
    print(xyz_loc)

    a = da.zeros(shape=(384, 128, 128), dtype=int)
    num_channels = 3
    a[1::num_channels, ...] = 1
    a[2::num_channels, ...] = 2
    channels = Deinterleave.deinterleave(a, num_channels, axis=0)
    for c in channels:
        print(c.sum().compute())
        print(c.shape)

    print(dict(zip(names, channels)))



import math
from typing import Tuple, Union, Iterator

import dask.array as da
import numpy as np


class DimensionsError(Exception):
    pass


def ensure_array_5d(
    arr: Union[np.ndarray, da.Array]
) -> Union[np.ndarray, da.Array]:
    """
    Checks that the array is 5D, adding singleton dimensions to the
    start of the array if less, throwing a DimensionsError if more
    Args:
        arr: the arraylike object
    Returns:
        the 5D array
    Raises:
        DimensionsError: if the array has more than 5 dimensions
    """
    if arr.ndim > 5:
        raise DimensionsError("Only arrays up to 5D are supported")
    while arr.ndim < 5:
        arr = arr[np.newaxis, ...]
    return arr


def ensure_shape_5d(shape: Tuple[int, ...]) -> Tuple[int, int, int, int, int]:
    """
    Checks that the shape is 5D, adding singleton dimensions to the
    start of the sequence if less, throwing a DimensionsError if more.
    Args:
        shape: the sequence
    Returns:
        the 5D tuple
    Raises:
        DimensionsError: if the input has more than 5 dimensions
    """
    if len(shape) > 5:
        raise DimensionsError("Only shapes up to 5D are supported")
    while len(shape) < 5:
        shape = (1, *shape)
    return shape


def guess_chunks(
    data_shape: Tuple[int, int, int],
    target_size: int,
    itemsize: int,
    mode: str = "z",
) -> Tuple[int, int, int]:
    """
    Given the shape of a 3D array, determine the optimal chunk shape
    closest to target_size.
    Args:
        data_shape: the shape of the input array
        target_size: target chunk size in bytes
        itemsize: the number of bytes per array element
        mode: chunking strategy. Must be one of "z", "cycle", or "iso"
    Returns:
        the optimal chunk shape
    """
    if any(s < 1 for s in data_shape):
        raise ValueError("data_shape must be >= 1 for all dimensions")
    if target_size <= 0:
        raise ValueError("target_size must be > 0")
    if itemsize <= 0:
        raise ValueError("itemsize must be > 0")
    if mode == "z":
        plane_size = data_shape[1] * data_shape[2] * itemsize
        nplanes_per_chunk = int(math.ceil(target_size / plane_size))
        nplanes_per_chunk = min(nplanes_per_chunk, data_shape[0])
        chunks = (
            nplanes_per_chunk,
            data_shape[1],
            data_shape[2],
        )
    elif mode == "cycle":
        current = np.array(data_shape, dtype=np.uint64)
        idx = 0
        ndims = len(current)
        while _get_size(current, itemsize) > target_size:
            current[idx % ndims] = int(math.ceil(current[idx % ndims] / 2.0))
            idx += 1
        chunks = (current[0], current[1], current[2])
    elif mode == "iso":
        chunk_dim = int(math.ceil((target_size / itemsize) ** (1.0 / 3)))
        chunks = (
            min(data_shape[0], chunk_dim),
            min(data_shape[1], chunk_dim),
            min(data_shape[2], chunk_dim),
        )
    else:
        raise ValueError(f"Invalid mode {mode}")

    # convert numpy int64 to Python int or zarr will complain
    return tuple(int(d) for d in chunks)


def expand_chunks(
    chunks: Tuple[int, int, int],
    data_shape: Tuple[int, int, int],
    target_size: int,
    itemsize: int,
    mode: str = "iso",
) -> Tuple[int, int, int]:
    """
    Given the shape and chunk size of a pre-chunked 3D array, determine the optimal chunk shape
    closest to target_size. Expanded chunk dimensions are an integer multiple of the base chunk dimension,
    to ensure optimal access patterns.
    Args:
        chunks: the shape of the input array chunks
        data_shape: the shape of the input array
        target_size: target chunk size in bytes
        itemsize: the number of bytes per array element
        mode: chunking strategy. Must be one of "cycle", or "iso"
    Returns:
        the optimal chunk shape
    """
    if any(c < 1 for c in chunks):
        raise ValueError("data_shape must be >= 1 for all dimensions")
    if any(s < 1 for s in data_shape):
        raise ValueError("data_shape must be >= 1 for all dimensions")
    if any(c > s for c, s in zip(chunks, data_shape)):
        raise ValueError(
            "chunks cannot be larger than data_shape in any dimension"
        )
    if target_size <= 0:
        raise ValueError("target_size must be > 0")
    if itemsize <= 0:
        raise ValueError("itemsize must be > 0")
    if mode == "cycle":
        # get the spatial dimensions only
        current = np.array(chunks, dtype=np.uint64)
        prev = current.copy()
        idx = 0
        ndims = len(current)
        while _get_size(current, itemsize) < target_size:
            prev = current.copy()
            current[idx % ndims] *= 2
            idx += 1
        current = _closer_to_target(current, prev, target_size, itemsize)
        expanded = (
            min(data_shape[0], current[0]),
            min(data_shape[1], current[1]),
            min(data_shape[2], current[2]),
        )
    elif mode == "iso":
        initial = np.array(chunks, dtype=np.uint64)
        current = initial
        prev = current
        i = 2
        while _get_size(current, itemsize) < target_size:
            prev = current
            current = initial * i
            i += 1
        current = _closer_to_target(current, prev, target_size, itemsize)
        expanded = (
            min(data_shape[0], current[0]),
            min(data_shape[1], current[1]),
            min(data_shape[2], current[2]),
        )
    else:
        raise ValueError(f"Invalid mode {mode}")

    return tuple(int(d) for d in expanded)


# https://stackoverflow.com/a/47518679
def range_with_end(start: int, end: int, step: int) -> Iterator[int]:
    """Generator for a range function which always includes the end value.
    Args:
        start: the start value
        end: the end value (inclusive)
        step: the step size
    Returns:
        A generator of integers
    Example:
        >>> list(range_with_end(0, 10, 3))
        >>> [0, 3, 6, 9, 10]
    """
    i = start
    while i < end:
        yield i
        i += step
    yield end


def _closer_to_target(
    shape1: Tuple[int, ...],
    shape2: Tuple[int, ...],
    target_bytes: int,
    itemsize: int,
) -> Tuple[int, ...]:
    """
    Given two shapes with the same number of dimensions,
    find which one is closer to target_bytes.
    Args:
        shape1: the first shape
        shape2: the second shape
        target_bytes: the target size for the returned shape
        itemsize: number of bytes per array element
    """
    size1 = _get_size(shape1, itemsize)
    size2 = _get_size(shape2, itemsize)
    if abs(size1 - target_bytes) < abs(size2 - target_bytes):
        return shape1
    return shape2


def _get_size(shape: Tuple[int, ...], itemsize: int) -> int:
    """
    Return the size of an array with the given shape, in bytes
    Args:
        shape: the shape of the array
        itemsize: number of bytes per array element
    Returns:
        the size of the array, in bytes
    """
    if any(s <= 0 for s in shape):
        raise ValueError("shape must be > 0 in all dimensions")
    return np.product(shape) * itemsize



import argparse
import errno
import itertools
import logging
import multiprocessing
import os
import subprocess
import time
from datetime import datetime
from pathlib import PurePath

import numpy as np
from google.cloud.storage import Blob

from aind_data_transfer.gcs import GCSUploader, create_client
from aind_data_transfer.util.dask_utils import (
    get_client,
    log_dashboard_address,
)
from aind_data_transfer.util.file_utils import (
    collect_filepaths,
    make_cloud_paths,
)

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


GCLOUD_TOOLS = ["gcloud alpha storage", "gsutil"]


def _chunk_files(filepaths, n_workers, tasks_per_worker):
    """
    Partition a list of filepaths into chunks.
    Our target is ~tasks_per_worker chunks of files for each dask worker to upload.
    The way each job (upload of a single chunk) is actually distributed is up to the scheduler.
    Some workers may get many jobs while others get few.
    Increasing tasks_per_worker can help the scheduler balance jobs among workers.
    Args:
        filepaths (list): list of absolute paths to chunk
        n_workers (int): the number of dask workers
        tasks_per_worker (int): target # tasks per dask worker
    Returns:
        a list of lists, where each sublist contains a list of filepaths
    """
    n = min(n_workers * tasks_per_worker, len(filepaths))
    chunked_files = np.array_split(filepaths, n)
    logger.info(
        f"Split files into {len(chunked_files)} chunks with "
        f"{len(chunked_files[0])} files each"
    )
    return chunked_files


def _python_upload_worker(
    bucket_name, files, gcs_path, root=None, chunk_size=256 * 1024 * 1024
):
    """
    Uploads a list of filepaths to GCS using the google-cloud-storage Python API
    Args:
        bucket_name (str): name of bucket
        files (list): list of absolute paths to upload to bucket
        gcs_path (str): google cloud storage path to upload files to
        root (str): directory shared by files to serve as root directory
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
    """
    uploader = GCSUploader(bucket_name)
    return uploader.upload_files(files, gcs_path, root, chunk_size=chunk_size)


def run_cluster_job(
    filepaths,
    bucket,
    gcs_path,
    tasks_per_worker,
    root=None,
    chunk_size=256 * 1024 * 1024,
):
    """
    Upload a list of filepaths using multiple distributed instances of the google-cloud-storage Python API
    Args:
        filepaths (list): list of absolute paths to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        tasks_per_worker (int): target number of chunks for each worker to process
        root (str): shared directory for filepaths to serve as root folder in the bucket
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
    """
    client, ntasks = get_client(deployment="slurm")
    chunked_files = _chunk_files(filepaths, ntasks, tasks_per_worker)
    futures = []
    for i, chunk in enumerate(chunked_files):
        futures.append(
            client.submit(
                _python_upload_worker,
                bucket_name=bucket,
                files=chunk,
                gcs_path=gcs_path,
                root=root,
                chunk_size=chunk_size,
            )
        )
    failed_uploads = list(itertools.chain(*client.gather(futures)))
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")
    client.shutdown()


def run_python_local_job(
    input_dir,
    bucket,
    gcs_path,
    n_workers=4,
    chunk_size=256 * 1024 * 1024,
    exclude_dirs=None,
):
    """
    Upload a directory using the google-cloud-storage Python API and multiprocessing
    Args:
        input_dir (str): directory to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        n_workers (int): number of workers
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
        exclude_dirs (list): list of directory names to exclude, e.g., ["dir1", "dir2"]
    """
    files = collect_filepaths(
        input_dir, recursive=True, exclude_dirs=exclude_dirs
    )
    chunked_files = _chunk_files(files, n_workers, tasks_per_worker=1)
    args = zip(
        itertools.repeat(bucket),
        chunked_files,
        itertools.repeat(gcs_path),
        itertools.repeat(input_dir),
        itertools.repeat(chunk_size),
    )
    with multiprocessing.Pool(n_workers) as pool:
        failed_uploads = list(
            itertools.chain(*pool.starmap(_python_upload_worker, args))
        )
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")


def _parse_cp_failures(log_path):
    """
    Parse the gsutil cp log for any transfers with a non-OK status code
    Args:
        log_path (str): path to the log file created by gsutil,
                        e.g., gsutil cp -L log_path.log ...
    Returns:
        a list of filepaths that resulted in an error during transfer
    """
    failures = []
    with open(log_path, "r") as f:
        next(f)  # skip header
        for line in f:
            parts = line.strip().split(",")
            src = parts[0]
            status = parts[-2]
            if status != "OK":
                failures.append(src)
    return failures


def _make_boto_options(n_threads):
    """Create Boto options to pass to gsutil
    See https://cloud.google.com/storage/docs/boto-gsutil
    Args:
        n_threads (int): number of threads to use for parallel uploads
    Returns:
        a list of option strings
    """
    options = []
    # Use the same process / thread ratio as the ~/.boto defaults
    # These defaults are different on Windows (1 process, many threads)
    processes = max(1, n_threads - 1)
    options.append(f"Boto:parallel_thread_count={n_threads}")
    options.append(f"Boto:parallel_process_count={processes}")
    return options


def build_gcloud_cmd(
    tool,
    func,
    input_dir,
    bucket,
    gcs_path,
    nthreads,
    recursive=True,
    logfile=None,
):
    if tool not in GCLOUD_TOOLS:
        raise ValueError(f"Invalid tool {tool}")
    cmd = f"{tool}"
    if tool == "gsutil" and nthreads > 1:
        sep = " -o "
        cmd += f" -m {sep}{sep.join(_make_boto_options(nthreads))}"
    cmd += f" {func}"
    if logfile is not None:
        cmd += f" -L {logfile}"
    if recursive:
        cmd += " -r"
    cmd += f" {input_dir} gs://{bucket}/{gcs_path}"
    logger.debug(f"Built command: {cmd}")
    return cmd


def run_gcloud_local_job(tool, input_dir, bucket, gcs_path, nthreads):
    """
    Upload a directory using the gcloud storage CLI
    Args:
        tool (str): either "gcloud alpha storage" or "gsutil"
        input_dir (str): directory to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        nthreads (int): num threads to use for upload. Only applies to gsutil
    """
    # assumes there will only be a single space, which should always be the case
    logfile = f"{tool.replace(' ', '_')}-log-{datetime.utcnow().strftime('%Y%m%d%H%M%S%f')}.log"
    cmd = build_gcloud_cmd(
        tool,
        func="cp",
        input_dir=input_dir,
        bucket=bucket,
        gcs_path=gcs_path,
        nthreads=nthreads,
        logfile=logfile,
    )
    ret = subprocess.run(cmd, shell=True)
    if ret.returncode != 0:
        logger.error(f"{tool} exited with code {ret.returncode}")
        failed_uploads = _parse_cp_failures(logfile)
        logger.error(
            f"{len(failed_uploads)} failed uploads:\n{failed_uploads}"
        )
    try:
        os.remove(logfile)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def validate_blobs(bucket_name, target_paths):
    """
    Ensure that all blobs were uploaded correctly
    Args:
        bucket_name (str): bucket name
        target_paths (list): list of cloud paths to validate
    Returns:
        list of cloud storage paths that are missing a Blob
    """
    storage_client = create_client()
    bucket = storage_client.get_bucket(bucket_name)
    missing_paths = []
    for path in target_paths:
        if not Blob(path, bucket).exists():
            missing_paths.append(path)
    return missing_paths


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="folder or file to upload",
    )
    parser.add_argument(
        "-b",
        "--bucket",
        type=str,
        help="s3 bucket",
    )
    parser.add_argument(
        "--gcs_path",
        type=str,
        default=None,
        help='gcs path relative to bucket, e.g., "folder/data.h5"',
    )
    parser.add_argument(
        "--recursive",
        default=False,
        action="store_true",
        help="upload a folder recursively",
    )
    parser.add_argument(
        "--tasks_per_worker",
        type=int,
        default=3,
        help="Split the work so that each worker has this many jobs. "
        "Increasing this can help with worker/task imbalance.",
    )
    parser.add_argument(
        "--cluster",
        default=False,
        action="store_true",
        help="run in cluster mode",
    )
    parser.add_argument(
        "--nthreads",
        type=int,
        default=1,
        help="number of threads if running locally",
    )
    parser.add_argument(
        "--chunk_size",
        type=int,
        default=256,
        help="upload files in chunks of this size (MB)",
    )
    parser.add_argument(
        "--method",
        choices=["python", "gsutil", "gcloud_alpha_storage"],
        default="python",
        help="use either gsutil, gcloud storage or the google-cloud-storage Python API for local upload."
        "This does not apply when --cluster is used.",
    )
    parser.add_argument(
        "--validate",
        default=False,
        action="store_true",
        help="Validate that all uploads were successful. This can take a while if there are many files."
        "This is an additional billed API request for each uploaded object.",
    )
    parser.add_argument(
        "--exclude_dirs",
        type=str,
        nargs="+",
        default=None,
        help="directories to exclude from upload",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    gcs_path = args.gcs_path
    if gcs_path is None:
        gcs_path = PurePath(args.input).name
    # remove leading slash since it will result in the structure
    # "bucket/'/'/s3_path" instead of "bucket/s3_path"
    gcs_path = gcs_path.strip("/")
    logger.info(f"Will upload to {args.bucket}/{gcs_path}")

    filepaths = collect_filepaths(
        args.input, recursive=args.recursive, exclude_dirs=args.exclude_dirs
    )
    cloud_paths = make_cloud_paths(filepaths, args.gcs_path, args.input)

    t0 = time.time()
    if args.cluster:
        chunk_size = args.chunk_size * 1024 * 1024  # MB to bytes
        run_cluster_job(
            filepaths=filepaths,
            bucket=args.bucket,
            gcs_path=args.gcs_path,
            tasks_per_worker=args.tasks_per_worker,
            root=args.input,
            chunk_size=chunk_size,
        )
    else:
        # replace CLI friendly string (no spaces) with actual gcloud command string
        norm_method = args.method.replace("_", " ")
        if norm_method in GCLOUD_TOOLS:
            # TODO: add exclusions
            run_gcloud_local_job(
                norm_method,
                args.input,
                args.bucket,
                args.gcs_path,
                args.nthreads,
            )
        elif norm_method == "python":
            run_python_local_job(
                args.input,
                args.bucket,
                args.gcs_path,
                args.nthreads,
                exclude_dirs=args.exclude_dirs,
            )
        else:
            raise ValueError(f"Unsupported method {norm_method}")

    logger.info(f"Upload done. Took {time.time() - t0}s ")

    if args.validate:
        missing_blobs = validate_blobs(
            bucket_name=args.bucket, target_paths=cloud_paths
        )
        if not missing_blobs:
            logger.info("All uploads were successful")
        else:
            logger.info(
                f"{len(missing_blobs)} blobs missing.\n{missing_blobs}"
            )


if __name__ == "__main__":
    main()



import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from numcodecs import Blosc
from wavpack_numcodecs import WavPack

from aind_data_transfer.readers.ephys_readers import EphysReaders, DataReader
from aind_data_transfer.transformations.ephys_compressors import (
    EphysCompressors, CompressorName
)
from aind_data_transfer.transformations.generic_compressors import (
    VideoCompressor,
    ZipCompressor,
)
from aind_data_transfer.transformations.imaging_compressors import (
    ImagingCompressors,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_DIR = TEST_DIR / "resources"


class TestEphysCompressors(unittest.TestCase):
    def test_get_compressor(self):
        blosc_configs = {
            "cname": "zstd",
            "clevel": 9,
            "shuffle": Blosc.BITSHUFFLE,
        }
        wavpack_configs = {"level": 3}
        blosc = EphysCompressors.get_compressor(
            CompressorName.BLOSC.value, **blosc_configs
        )
        wavpack = EphysCompressors.get_compressor(
            CompressorName.WAVPACK.value, **wavpack_configs
        )
        expected_blosc = Blosc(
            cname="zstd", clevel=9, shuffle=Blosc.BITSHUFFLE
        )
        expected_wavpack = WavPack(level=3)
        self.assertEqual(blosc, expected_blosc)
        self.assertEqual(wavpack, expected_wavpack)

    def test_get_compressor_fails(self):
        with self.assertRaises(Exception):
            EphysCompressors.get_compressor("Made up name")

    def test_scale_recording(self):
        open_ephys_dir = (
            RESOURCES_DIR / "v0.6.x_neuropixels_multiexp_multistream"
        )
        openephys_reader = DataReader.OPENEPHYS.value
        read_blocks = EphysReaders.get_read_blocks(
            openephys_reader, open_ephys_dir
        )
        read_block = next(read_blocks)
        chunk_size = min(read_block["recording"].get_num_frames(0) - 1, 10000)
        # The first read block should be a NI-DAQ recording.
        self.assertTrue(
            EphysReaders.RecordingBlockPrefixes.nidaq.value
            in read_block["stream_name"],
            "The first block isn't NI-DAQ. If the test data has "
            "been updated, this test needs to be updated too.",
        )
        # We expect the NI-DAQ scaled recordings to be the same as the original
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            [read_block], chunk_size=chunk_size
        )
        scaled_read_block = next(scaled_read_blocks)
        self.assertEqual(
            scaled_read_block["scaled_recording"], read_block["recording"]
        )

        read_block = next(read_blocks)
        # The second read block should be a Neuropixel recording.
        self.assertTrue(
            EphysReaders.RecordingBlockPrefixes.neuropix.value
            in read_block["stream_name"],
            "The second block isn't Neuropixel. If the test data "
            "has been updated, this test needs to be updated too.",
        )
        chunk_size = min(read_block["recording"].get_num_frames(0) - 1, 10000)
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            [read_block], chunk_size=chunk_size
        )
        scaled_read_block = next(scaled_read_blocks)


class TestImagingCompressors(unittest.TestCase):
    def test_get_compressor(self):
        blosc_configs = {
            "cname": "zstd",
            "clevel": 1,
            "shuffle": Blosc.SHUFFLE,
        }
        blosc = ImagingCompressors.get_compressor(
            ImagingCompressors.Compressors.blosc.name, **blosc_configs
        )
        expected_blosc = Blosc(cname="zstd", clevel=1, shuffle=Blosc.SHUFFLE)
        self.assertEqual(blosc, expected_blosc)

    def test_get_compressor_fails(self):
        with self.assertRaises(Exception):
            ImagingCompressors.get_compressor("Made up name")


class TestGenericCompressor(unittest.TestCase):
    """Tests VideoCompressor class methods"""

    def test_video_compressor_creation(self):
        """Tests VideoCompressor default is set correctly"""
        video_compressor = VideoCompressor()
        self.assertEqual(video_compressor.compression_level, 5)

    @patch("pyminizip.compress_multiple")
    def test_zip_compressor(self, mock_zip: MagicMock):
        """Tests zip compression. No subdirectories or files skipped."""
        zc = ZipCompressor()
        input_dir = (
            RESOURCES_DIR
            / "v0.6.x_neuropixels_multiexp_multistream"
            / "Videos"
        )
        output_dir = RESOURCES_DIR
        file_names = [
            str(input_dir / "a_video_file.avi"),
            str(input_dir / "another_video_file.avi"),
            str(input_dir / "video_ref.csv"),
        ]
        prefix_names = [str("Videos"), str("Videos"), str("Videos")]
        zc.compress_dir(input_dir=input_dir, output_dir=output_dir)
        file_names_call = mock_zip.mock_calls[0].args[0]
        prefix_names_call = mock_zip.mock_calls[0].args[1]
        zipped_actual_calls = zip(file_names_call, prefix_names_call)
        zipped_expected_calls = zip(file_names, prefix_names)
        output_dir_call = mock_zip.mock_calls[0].args[2]
        self.assertEqual(output_dir_call, str(RESOURCES_DIR))
        self.assertEqual(set(zipped_expected_calls), set(zipped_actual_calls))

    @patch("pyminizip.compress_multiple")
    def test_zip_compressor_skip_dir(self, mock_zip: MagicMock):
        """Tests zip compression. One file skipped."""
        zc = ZipCompressor()
        input_dir = (
            RESOURCES_DIR
            / "v0.6.x_neuropixels_multiexp_multistream"
            / "Videos"
        )
        skip_file = (
            RESOURCES_DIR
            / "v0.6.x_neuropixels_multiexp_multistream"
            / "Videos"
            / "a_video_file.avi"
        )
        output_dir = RESOURCES_DIR
        file_names = [
            str(input_dir / "another_video_file.avi"),
            str(input_dir / "video_ref.csv"),
        ]
        prefix_names = [str("Videos"), str("Videos")]
        zc.compress_dir(
            input_dir=input_dir, output_dir=output_dir, skip_dirs=[skip_file]
        )
        file_names_call = mock_zip.mock_calls[0].args[0]
        prefix_names_call = mock_zip.mock_calls[0].args[1]
        zipped_actual_calls = zip(file_names_call, prefix_names_call)
        zipped_expected_calls = zip(file_names, prefix_names)
        output_dir_call = mock_zip.mock_calls[0].args[2]
        self.assertEqual(output_dir_call, str(RESOURCES_DIR))
        self.assertEqual(set(zipped_expected_calls), set(zipped_actual_calls))


if __name__ == "__main__":
    unittest.main()



from datetime import datetime
import xml.etree.ElementTree as ET
import pandas as pd

import file_io

from aind_data_schema.imaging.acquisition import AxisName, Direction, Axis, Immersion, Acquisition, AcquisitionTile
from aind_data_schema.imaging.tile import Channel, Scale3dTransform, Translation3dTransform

def log_to_acq_json(log_dict: dict) -> Acquisition:
    """
    Parameters
    ----------
    log_dict: dict
        Output of file_io.read_log_file. 
        log_dict formats general metadata as flat dictionary, and tile metadata in a nested dict. 
    Returns 
    -------
    Acquisition
        Acquisition instance
    """

    experimenter_full_name: str = 'ISpim Group'
    specimen_id: str = log_dict['specimen_id']
    subject_id: str = log_dict['subject_id']
    instrument_id: str = log_dict['instrument_id']
    session_start_time: datetime = log_dict['session_start_time']
    session_end_time: datetime = log_dict['session_end_time']

    tiles: list[AcquisitionTile] = []
    for tile_dict in log_dict['tiles']:
        ch = Channel(channel_name=tile_dict['channel_name'], 
                     laser_wavelength=int(tile_dict['laser_wavelength']), 
                     laser_power=tile_dict['laser_power'],
                     filter_wheel_index=tile_dict['filter_wheel_index'])
        scale_tfm = Scale3dTransform(scale=[float(tile_dict['x_voxel_size']), 
                                            float(tile_dict['y_voxel_size']), 
                                            float(tile_dict['z_voxel_size'])])
        translation_tfm = Translation3dTransform(translation=
                        [float(tile_dict['tile_x_position']) * 1000. / float(tile_dict['x_voxel_size']), 
                        float(tile_dict['tile_y_position']) * 1000. / float(tile_dict['y_voxel_size']),
                        float(tile_dict['tile_z_position']) * 1000. / float(tile_dict['z_voxel_size'])])
        tile = AcquisitionTile(channel=ch, 
                               file_name=tile_dict['file_name'],
                               imaging_angle=tile_dict['lightsheet_angle'], 
                               coordinate_transformations=[scale_tfm, translation_tfm])
        tiles.append(tile)


    # NOTE: Made up directions
    axes: list[Axis] = []
    axes.append(Axis(name=AxisName.X, 
                     dimension=2, 
                     direction=Direction.LR))  
    axes.append(Axis(name=AxisName.Y, 
                     dimension=1, 
                     direction=Direction.AP))
    axes.append(Axis(name=AxisName.Z, 
                     dimension=0, 
                     direction=Direction.IS))

    chamber_immersion: Immersion = Immersion(medium=log_dict['chamber_immersion_medium'], 
                                             refractive_index=log_dict['chamber_immersion_refractive_index'])
    local_storage_directory: str = log_dict['local_storage_directory']
    external_storage_directory: str = log_dict['external_storage_directory']

    return Acquisition(experimenter_full_name=experimenter_full_name, 
                       specimen_id=specimen_id, 
                       subject_id=subject_id, 
                       instrument_id=instrument_id, 
                       session_start_time=session_start_time, 
                       session_end_time=session_end_time, 
                       tiles=tiles, 
                       axes=axes,
                       chamber_immersion=chamber_immersion, 
                       local_storage_directory=local_storage_directory,
                       external_storage_directory=external_storage_directory)

def acq_json_to_xml(acq_obj: Acquisition, s3_path: str, zarr: bool = True, condition: str = "") -> ET.ElementTree:
    """
    Parameters
    ----------
    acq_obj: Acquisition
        Acquisition instance
    s3_path: str
        S3 URI to dataset
    zarr: bool
        Boolean that describes the type of the dataset: zarr or n5. 
        If n5, conditioning the dataset is turned off. 
    condition: str
        Conditional statement for filtering the dataset on specific attributes.
        Can filter on following 5 fields: X, Y, Z, channel, camera
        Ex: "channel == 405 and camera = 0"    

    Returns
    -------
    ET.ElementTree: 
        ET.ElementTree instance
    """
    
    # Retrieve map of tile names -> transforms
    def extract_tile_translation() -> dict[str, list[float]]:
        tile_transforms: dict[str, list[float]] = {}

        for tile in acq_obj['tiles']:
            translation: list[float] = tile['coordinate_transformations'][1]['translation']
            filename: str = tile['file_name']

            tile_transforms[filename] = translation

        return tile_transforms
    
    # Parses tile names into a pandas dataframe and 
    # applies boolean masking on given str condition. 
    # Returns list of filtered tile names. 
    def filter_tiles_on_condition(condition: str) -> list[str]:
        """
        Condition Field Names: 
        X, Y, Z, channel, camera
        """

        # Construct Dataframe
        fields = {'filename': '', 
                  'X': int(), 
                  'Y': int(),
                  'Z': int(),
                  'channel': int(),
                  'camera': int()}
        df = pd.DataFrame(fields, index=[])
        for tile in acq_obj['tiles']: 
            filename: str = tile['file_name']
            vals = filename.split('_')
            X = int(vals[2])
            Y = int(vals[4])
            Z = int(vals[6])
            
            if len(vals[8]) <= 4: 
                channel = int(vals[8])
                cam = int(vals[9][-1])
            else:
                channel = int(vals[8][:vals[8].index('.')])
                cam = -1

            df.append({'X': X, 
                       'Y': Y, 
                       'Z': Z,
                       'channel': channel,
                       'cam': cam}, ignore_index=True)

        # Filter Dataframe
        if condition != "": 
            results = df.query(condition)
        else:
            results = df
        output_tilenames = list(results['filename'])

        return output_tilenames

    # Define filepaths of tiles and associated attributes
    def add_sequence_description(parent: ET.Element, filtered_tiles: list[str]) -> None: 
        def add_image_loader(seq_desc: ET.Element, filtered_tiles: list[str]) -> None:
            img_loader = ET.SubElement(seq_desc, "ImageLoader")
            if zarr: 
                img_loader.attrib["format"] = "bdv.multimg.zarr"
                img_loader.attrib["version"] = "1.0"
                x = ET.SubElement(img_loader, "zarr")
                x.attrib["type"] = "absolute"

                # FIXME: Test later
                x.text = 'home/TODO_ADD_USER/' + s3_path[s3_path.index('//') + 2:]

                zgs = ET.SubElement(img_loader, "zgroups")
                for i, tile in enumerate(filtered_tiles):
                    zg = ET.SubElement(zgs, "zgroup")
                    zg.attrib["setup"] = f"{i}"
                    zg.attrib["timepoint"] = "0"
                    x = ET.SubElement(zg, "path")
                    x.text = tile
                    
            else:  # n5 
                img_loader.attrib["format"] = "bdv.n5"
                img_loader.attrib["version"] = "1.0"
                x = ET.SubElement(img_loader, "n5")
                x.attrib["type"] = "relative"
                x.text = "dataset.n5"

        def add_view_setups(seq_desc: ET.Element, filtered_tiles: list[str]) -> None: 
            def add_attributes(view_setups: ET.Element) -> None:
                # Add attributes
                x = ET.SubElement(view_setups, "Attributes")
                x.attrib["name"] = "illumination"
                x = ET.SubElement(x, "Illumination")
                y = ET.SubElement(x, "id")
                y.text = "0"
                y = ET.SubElement(x, "name")
                y.text = "0"

                x = ET.SubElement(view_setups, "Attributes")
                x.attrib["name"] = "channel"
                x = ET.SubElement(x, "Channel")
                y = ET.SubElement(x, "id")
                y.text = "0"
                y = ET.SubElement(x, "name")
                y.text = "0"

                tile_atts = ET.SubElement(view_setups, "Attributes")
                tile_atts.attrib["name"] = "tile"
                for i, tile in enumerate(filtered_tiles):
                    t_entry = ET.Element("Tile")
                    id_entry = ET.Element("id")
                    id_entry.text = f"{i}"
                    t_entry.append(id_entry)
                    name_entry = ET.Element("name")
                    name_entry.text = tile
                    t_entry.append(name_entry)
                    tile_atts.append(t_entry)

                x = ET.SubElement(view_setups, "Attributes")
                x.attrib["name"] = "angle"
                x = ET.SubElement(x, "Angle")
                y = ET.SubElement(x, "id")
                y.text = "0"
                y = ET.SubElement(x, "name")
                y.text = "0"
            
            view_setups = ET.SubElement(seq_desc, "ViewSetups")
            for i, tile in enumerate(filtered_tiles):
                vs = ET.SubElement(view_setups, "ViewSetup")

                x = ET.SubElement(vs, "id")
                x.text = f"{i}"
                x = ET.SubElement(vs, "name")
                x.text = tile
                x = ET.SubElement(vs, "size")
                
                # FIXME: Test later
                # metadata_path_res = 'home/TODO_ADD_USER/' + s3_path[s3_path.index('//') + 2:] + '0/.zarray'
                # shape: list[int] = file_io.read_json(metadata_path_res)["shape"]  # 5 ints: [t, c, z, y, x]
                # x.text = f"{shape[4]} {shape[3]} {shape[2]}"  
                x.text = f"TODO TODO TODO"

                voxel_size = ET.SubElement(vs, "voxelSize")
                x = ET.SubElement(voxel_size, "unit")
                x.text = "µm"
                x = ET.SubElement(voxel_size, "size")

                scale_vector = acq_obj["tiles"][0]['coordinate_transformations'][0]['scale']
                x.text = f"{scale_vector[0]} {scale_vector[1]} {scale_vector[2]}"

                attr = ET.SubElement(vs, "attributes")
                x = ET.SubElement(attr, "illumination")
                x.text = "0"
                x = ET.SubElement(attr, "channel")
                x.text = "0"
                x = ET.SubElement(attr, "tile")
                x.text = f"{i}"
                x = ET.SubElement(attr, "angle")
                x.text = "0"

            add_attributes(view_setups)

        def add_time_points(seq_desc: ET.Element) -> None: 
            x = ET.SubElement(seq_desc, "Timepoints")
            x.attrib["type"] = "pattern"
            y = ET.SubElement(x, "integerpattern")
            y.text = "0"
            ET.SubElement(seq_desc, "MissingViews")

        seq_desc = ET.SubElement(parent, "SequenceDescription")
        add_image_loader(seq_desc, filtered_tiles)
        add_view_setups(seq_desc, filtered_tiles)
        add_time_points(seq_desc)

    # Define transformations applied to tiles
    def add_view_registrations(parent: ET.Element, filtered_translations: list[list[float]]) -> None:
        view_registrations = ET.SubElement(parent, "ViewRegistrations")
        for i, tr in enumerate(filtered_translations): 
            vr = ET.SubElement(view_registrations, "ViewRegistration")
            vr.attrib["timepoint"] = "0"
            vr.attrib["setup"] = f"{i}"

            vt = ET.Element("ViewTransform")
            vt.attrib["type"] = "affine"
            name = ET.SubElement(vt, "Name")
            name.text = "Translation to Nominal Grid"
            affine = ET.SubElement(vt, "affine")
            affine.text = f'0.0 0.0 0.0 {tr[0]} 0.0 0.0 0.0 {tr[1]} 0.0 0.0 0.0 {tr[2]}'

            vr.append(vt)

    # Gather info from acq json
    tile_translations: dict[str, list[float]] = extract_tile_translation()
    filtered_tiles: list[str] = list(tile_translations.keys())
    filtered_translations: list[list[float]] = list(tile_translations.values())
    if zarr and condition != "":
        filtered_tiles = filter_tiles_on_condition(condition)
        filtered_translations = tile_translations[filtered_tiles]

    # Construct the output xml
    spim_data = ET.Element("SpimData")
    spim_data.attrib["version"] = "0.2"
    x = ET.SubElement(spim_data, "BasePath")
    x.attrib["type"] = "relative"
    x.text = "."
    add_sequence_description(spim_data, filtered_tiles)
    add_view_registrations(spim_data, filtered_translations)

    return ET.ElementTree(spim_data)


import fnmatch
import logging
import time
from pathlib import Path
from typing import List, Optional, Dict, cast

import dask
import dask.array
import tifffile
import zarr
from distributed import wait
from numpy.typing import NDArray
from ome_zarr.format import CurrentFormat
from ome_zarr.io import parse_url
from ome_zarr.writer import write_multiscale
from xarray_multiscale import multiscale
from xarray_multiscale.reducers import windowed_mean

from aind_data_transfer.util.chunk_utils import *
from aind_data_transfer.util.file_utils import collect_filepaths
from aind_data_transfer.util.io_utils import (
    DataReaderFactory,
    ImarisReader,
    MissingDatasetError,
    DataReader
)
from aind_data_transfer.transformations.deinterleave import ChannelParser, Deinterleave
from aind_data_transfer.transformations.flatfield_correction import BkgSubtraction

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def write_files(
    image_paths: list,
    output: str,
    n_levels: int,
    scale_factor: int,
    overwrite: bool = True,
    chunk_size: float = 64,  # MB
    chunk_shape: tuple = None,
    voxel_size: tuple = None,
    storage_options: dict = None,
    bkg_img_dir: Optional[str] = None
) -> list:
    """
    Write each image as a separate group to a Zarr store.
    Each group will be named according to the image filename minus extension.
    For example, if a file has path /data/tiff/tile_0_0.tif, the group name
    for that image is tile_0_0. Within each group there are n_levels arrays,
    one for each resolution level.
    Args:
        image_paths: the list of image filepaths to convert to Zarr
        output: the location of the output Zarr store (filesystem, s3, gs)
        n_levels: number of downsampling levels
        scale_factor: scale factor for downsampling in X, Y and Z
        overwrite: whether to overwrite image groups that already exist
        chunk_size: the target chunk size in MB
        chunk_shape: the chunk shape, if None will be computed from chunk_size
        voxel_size: three element tuple giving physical voxel sizes in Z,Y,X order
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
        bkg_img_dir: the directory containing the background image Tiff file for each raw image.
                     If None, will not perform background subtraction.
    Returns:
        A list of metrics for each converted image
    """
    if chunk_shape is None and chunk_size <= 0:
        raise ValueError(
            "Either chunk_shape must be set or chunk_size must be greater than 0"
        )
    if chunk_shape is not None and any(s < 1 for s in chunk_shape):
        raise ValueError("chunk_shape must be at least 1 in all dimensions")

    if not image_paths:
        LOGGER.warning("No images found. Exiting.")
        return []

    if storage_options is None:
        storage_options = {}

    store = parse_url(output, mode="w").store
    root_group = zarr.group(store=store)

    all_metrics = []

    for impath in sorted(image_paths):
        LOGGER.info(f"Writing tile {impath}")

        # Create readers, but don't construct dask array
        # until we know the optimal chunk shape.
        with DataReaderFactory().create(impath) as reader:
            # TODO: pass units to zarr writer
            if voxel_size is None:
                try:
                    voxel_size, _ = reader.get_voxel_size()
                except Exception:
                    voxel_size = [1.0, 1.0, 1.0]
            LOGGER.info(f"Using voxel size: {voxel_size}")

            # Attempt to parse tile origin
            origin = _parse_origin(reader)

            if chunk_shape is None:
                chunks = _compute_chunks(reader, chunk_size)
            else:
                chunks = tuple(chunk_shape)

            LOGGER.info(
                f"chunks: {chunks}, {np.product(chunks) * reader.get_itemsize() / (1024 ** 2)} MiB"
            )

            try:
                channel_names = ChannelParser.parse_channel_names(impath)
                LOGGER.info(f"parsed channel wavelengths: {channel_names}")
                interleaved = len(channel_names) > 1
            except ValueError:
                interleaved = False
            LOGGER.info(f"Data is interleaved: {interleaved}")

            if interleaved:
                LOGGER.info("Deinterleaving during write")
                tile_metrics = _store_interleaved_file(
                    reader,
                    root_group,
                    output,
                    channel_names,
                    n_levels,
                    scale_factor,
                    origin,
                    voxel_size,
                    chunks,
                    overwrite,
                    storage_options
                )
                all_metrics.append(tile_metrics)
            else:
                tile_metrics = _store_file(
                    reader,
                    root_group,
                    output,
                    n_levels,
                    scale_factor,
                    origin,
                    voxel_size,
                    chunks,
                    overwrite,
                    storage_options,
                    bkg_img_dir=bkg_img_dir
                )
                all_metrics.append(tile_metrics)

    return all_metrics


def _store_file(
        reader: DataReader,
        root_group: zarr.Group,
        output: str,
        n_levels: int,
        scale_factor: int,
        origin: list,
        voxel_size: tuple,
        chunks: tuple,
        overwrite: bool,
        storage_options: dict,
        bkg_img_dir: Optional[str] = None
) -> dict:
    """
    Write an image to an existing Zarr store

    Args:
        reader: the reader for the image
        writer: the OmeZarrWriter instance
        output: the location of the output Zarr store (filesystem, s3, gs)
        n_levels: number of downsampling levels
        scale_factor: scale factor for downsampling in X, Y and Z
        origin: the tile origin coordinates in ZYX order
        physical_pixel_sizes: the voxel size of the image
        chunks: the chunk shape
        overwrite: whether to overwrite image groups that already exist
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
        bkg_img_dir: the directory containing the background image Tiff file for each raw image.
                     If None, will not perform background subtraction.
    Returns:
        A list of metrics for each converted image
    """
    tile_name = Path(reader.get_filepath()).stem + ".zarr"
    LOGGER.info(f"new tile name: {tile_name}")

    tile_metrics = {"chunks": chunks, "tile": tile_name}
    try:
        tile_metrics.update(storage_options["params"])
    except KeyError:
        LOGGER.warning("compression parameters will not be logged")

    if not overwrite and _tile_exists(output, tile_name, n_levels):
        LOGGER.info(f"Skipping tile {tile_name}, already exists.")
        return tile_metrics

    # Get the chunk dimensions that exist in the original, un-padded image
    reader_chunks = chunks[len(chunks) - len(reader.get_shape()):]

    pyramid = _get_or_create_pyramid(reader, n_levels, reader_chunks)

    if bkg_img_dir is not None:
        bkg_img_pyramid = _create_pyramid(
            tifffile.imread(BkgSubtraction.get_bkg_path(reader.get_filepath(), bkg_img_dir)),
            n_levels
        )
        for i in range(len(bkg_img_pyramid)):
            pyramid[i] = BkgSubtraction.subtract(
                pyramid[i],
                da.from_array(bkg_img_pyramid[i], chunks=(256, 256))
            )

    pyramid = [ensure_array_5d(arr) for arr in pyramid]

    LOGGER.info(f"{pyramid[0]}")

    ome_json = _build_ome(
        pyramid[0].shape,
        tile_name,
        channel_names=None,
        channel_colors=None,
        channel_minmax=None
    )
    axes_5d = _get_axes_5d()
    transforms, chunk_opts = _compute_scales(
        len(pyramid),
        (scale_factor,) * 3,
        voxel_size,
        chunks,
        pyramid[0].shape,
        origin,
    )
    for opt in chunk_opts:
        opt.update(storage_options)

    group = root_group.create_group(tile_name, overwrite=True)
    group.attrs["omero"] = ome_json

    LOGGER.info("Starting write...")
    t0 = time.time()
    jobs = write_multiscale(
        pyramid,
        group=group,
        fmt=CurrentFormat(),
        axes=axes_5d,
        coordinate_transformations=transforms,
        storage_options=chunk_opts,
        name=None,
        compute=False,
    )
    if jobs:
        LOGGER.info("Computing dask arrays...")
        arrs = dask.persist(*jobs)
        wait(arrs)
    write_time = time.time() - t0

    _populate_metrics(
        tile_metrics,
        tile_name,
        output,
        _get_bytes(pyramid),
        write_time,
        n_levels,
        pyramid[0].shape,
        pyramid[0].dtype,
    )

    LOGGER.info(
        f"Finished writing tile {tile_name}.\n"
        f"Took {write_time}s. {tile_metrics['write_bps'] / (1024 ** 2)} MiB/s"
    )

    return tile_metrics


def _store_interleaved_file(
        reader: DataReader,
        root_group: zarr.Group,
        output: str,
        channel_names: list,
        n_levels: int,
        scale_factor: int,
        origin: list,
        voxel_size: tuple,
        chunks: tuple,
        overwrite: bool,
        storage_options: dict,
) -> dict:
    """
    Write an image to an existing Zarr store

    Args:
        reader: the reader for the image
        writer: the OmeZarrWriter instance
        output: the location of the output Zarr store (filesystem, s3, gs)
        channel_names: the wavelength string of each channel, e.g., ["488, "561","689"]
        n_levels: number of downsampling levels to compute
        scale_factor: scale factor for downsampling in X, Y and Z
        origin: the tile origin coordinates in ZYX order
        physical_pixel_sizes: the voxel size of the image
        chunks: the chunk shape
        overwrite: whether to overwrite image groups that already exist
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
    Returns:
        A list of metrics for each converted image
    """
    tile_metrics = {"chunks": chunks}
    try:
        tile_metrics.update(storage_options["params"])
    except KeyError:
        LOGGER.warning("compression parameters will not be logged")

    # Get the chunk dimensions that exist in the original, un-padded image
    reader_chunks = chunks[len(chunks) - len(reader.get_shape()):]
    channels = Deinterleave.deinterleave(
        reader.as_dask_array(reader_chunks),
        len(channel_names),
        axis=0
    )
    for channel_idx, channel in enumerate(channels):
        tile_prefix = ChannelParser.parse_tile_xyz_loc(reader.get_filepath())
        tile_name = tile_prefix + f"_ch_{channel_names[channel_idx]}" + ".zarr"
        LOGGER.info(f"new tile name: {tile_name}")

        if not overwrite and _tile_exists(output, tile_name, n_levels):
            LOGGER.info(f"Skipping tile {tile_name}, already exists.")
            continue

        tile_metrics["tile"] = tile_name

        channel_pyramid = _create_pyramid(channel, n_levels, chunks=reader_chunks)
        channel_pyramid = [ensure_array_5d(arr) for arr in channel_pyramid]

        LOGGER.info(f"{channel_pyramid[0]}")

        ome_json = _build_ome(
            channel_pyramid[0].shape,
            tile_name,
            channel_names=None,
            channel_colors=None,
            channel_minmax=None
        )
        axes_5d = _get_axes_5d()
        transforms, chunk_opts = _compute_scales(
            len(channel_pyramid),
            (scale_factor,) * 3,
            voxel_size,
            chunks,
            channel_pyramid[0].shape,
            origin,
        )
        for opt in chunk_opts:
            opt.update(storage_options)

        group = root_group.create_group(tile_name, overwrite=True)
        group.attrs["omero"] = ome_json

        LOGGER.info("Starting write...")
        t0 = time.time()
        jobs = write_multiscale(
            channel_pyramid,
            group=group,
            fmt=CurrentFormat(),
            axes=axes_5d,
            coordinate_transformations=transforms,
            storage_options=chunk_opts,
            name=None,
            compute=False,
        )
        if jobs:
            LOGGER.info("Computing dask arrays...")
            arrs = dask.persist(*jobs)
            wait(arrs)
        write_time = time.time() - t0

        _populate_metrics(
            tile_metrics,
            tile_name,
            output,
            _get_bytes(channel_pyramid),
            write_time,
            n_levels,
            channel_pyramid[0].shape,
            channel_pyramid[0].dtype,
        )

        LOGGER.info(
            f"Finished writing tile {tile_name}.\n"
            f"Took {write_time}s. {tile_metrics['write_bps'] / (1024 ** 2)} MiB/s"
        )

    return tile_metrics


def write_folder(
    input: str,
    output: str,
    n_levels: int,
    scale_factor: int,
    overwrite: bool = True,
    chunk_size: float = 64,  # MB
    chunk_shape: tuple = None,
    voxel_size: tuple = None,
    exclude: list = None,
    storage_options: dict = None,
    recursive: bool = False,
    bkg_img_dir: Optional[str] = None
) -> list:
    """
    Write each image in the input directory as a separate group
    to a Zarr store. Each group will be named according to the image filename
    minus extension. For example, if a file has path /data/tiff/tile_0_0.tif, the group name
    for that image is tile_0_0. Within each group there are n_levels arrays, one for each resolution level.
    Args:
        input: the directory of images to convert to Zarr
        output: the location of the output Zarr store (filesystem, s3, gs)
        n_levels: number of downsampling levels
        scale_factor: scale factor for downsampling in X, Y and Z
        overwrite: whether to overwrite image groups that already exist
        chunk_size: the target chunk size in MB
        chunk_shape: the chunk shape, if None will be computed from chunk_size
        voxel_size: three element tuple giving physical voxel sizes in Z,Y,X order
        exclude: a list of filename patterns to exclude from conversion
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
        recursive: whether to convert all images in all subfolders
        bkg_img_dir: the directory containing the background image Tiff file for each raw image.
                     If None, will not perform background subtraction.
    Returns:
        A list of metrics for each converted image
    """
    if exclude is None:
        exclude = []

    image_paths = collect_filepaths(
        input,
        recursive=recursive,
        include_exts=DataReaderFactory().VALID_EXTENSIONS,
    )

    exclude_paths = set()
    for path in image_paths:
        if any(fnmatch.fnmatch(path, pattern) for pattern in exclude):
            exclude_paths.add(path)

    image_paths = [p for p in image_paths if p not in exclude_paths]
    LOGGER.info(f"Found {len(image_paths)} images to process")

    return write_files(
        image_paths,
        output,
        n_levels,
        scale_factor,
        overwrite,
        chunk_size,
        chunk_shape,
        voxel_size,
        storage_options,
        bkg_img_dir=bkg_img_dir
    )


def _parse_origin(reader):
    origin = None
    try:
        origin = reader.get_origin()
        while len(origin) < 5:
            origin = (0, *origin)
    except AttributeError:
        LOGGER.warning("Origin metadata could not be parsed from image")
    LOGGER.info(f"Using origin: {origin}")
    return origin


def _tile_exists(zarr_path, tile_name, n_levels):
    z = zarr.open(zarr_path, "r")
    try:
        # TODO: only re-upload missing levels
        a = z[f"{tile_name}/{n_levels - 1}"]
        return a.nbytes > 0
    except KeyError:
        return False


def _compute_chunks(reader, target_size_mb):
    target_size_bytes = target_size_mb * 1024 * 1024
    padded_chunks = ensure_shape_5d(reader.get_chunks())
    padded_shape = ensure_shape_5d(reader.get_shape())
    spatial_chunks = padded_chunks[2:]
    spatial_shape = padded_shape[2:]
    LOGGER.info(f"Using multiple of base chunk size: {padded_chunks}")
    if spatial_chunks[1:] == spatial_shape[1:]:
        LOGGER.info(
            "chunks and shape have same XY dimensions, "
            "will chunk along Z only."
        )
        chunks = guess_chunks(
            spatial_shape, target_size_bytes, reader.get_itemsize(), mode="z"
        )
    else:
        chunks = expand_chunks(
            spatial_chunks,
            spatial_shape,
            target_size_bytes,
            reader.get_itemsize(),
            mode="cycle",
        )
    chunks = ensure_shape_5d(chunks)
    return chunks


def _create_pyramid(
    data: Union[NDArray, dask.array.Array], n_lvls: int, chunks: Union[str, tuple] = "preserve"
) -> List[Union[NDArray, dask.array.Array]]:
    """
    Create a lazy multiscale image pyramid using data as the full-resolution layer.
    To evaluate the result, call dask.compute(*pyramid)

    Args:
        data: the numpy or dask array to create a pyramid from.
        This will be used as the highest resolution layer.
        n_lvls: the number of pyramid levels to produce
    Returns:
        A list of dask or numpy arrays, ordered from highest resolution to lowest
    """
    pyramid = multiscale(
        data,
        windowed_mean,  # func
        (2,) * data.ndim,  # scale factors
        preserve_dtype=True,
        chunks=chunks
    )[:n_lvls]
    return [arr.data for arr in pyramid]


def _get_or_create_pyramid(reader, n_levels, chunks):
    if isinstance(reader, ImarisReader):
        try:
            pyramid = reader.get_dask_pyramid(
                n_levels,
                timepoint=0,
                channel=0,
                # Use only the dimensions that exist in the base image.
                chunks=chunks,
            )
        except MissingDatasetError as e:
            LOGGER.error(e)
            LOGGER.warning(
                f"{reader.get_filepath()} does not contain all requested scales."
                f"Computing them instead..."
            )
            pyramid = _create_pyramid(
                reader.as_dask_array(chunks=chunks), n_levels, chunks
            )
    else:
        pyramid = _create_pyramid(
            reader.as_dask_array(chunks=chunks), n_levels, chunks
        )

    return pyramid


def _get_bytes(data: Union[List[NDArray], NDArray]):
    if isinstance(data, list):
        total_bytes = 0
        for arr in data:
            total_bytes += arr.nbytes
        return total_bytes
    return data.nbytes


def _get_storage_ratio(zarr_path: str, dataset_name: str):
    z = zarr.open(zarr_path, "r")
    full_res = z[f"{dataset_name}/0"]
    return full_res.nbytes / full_res.nbytes_stored


def _get_bytes_stored(zarr_path: str, dataset_name: str, n_levels: int):
    z = zarr.open(zarr_path, "r")
    total_bytes_stored = 0
    for res in range(n_levels):
        arr = z[f"{dataset_name}/{res}"]
        total_bytes_stored += arr.nbytes_stored
    return total_bytes_stored


def _populate_metrics(
    tile_metrics,
    tile_name,
    out_zarr,
    bytes_read,
    write_time,
    n_levels,
    shape,
    dtype,
):
    tile_metrics["n_levels"] = n_levels
    tile_metrics["shape"] = shape
    tile_metrics["dtype"] = dtype
    tile_metrics["write_time"] = write_time
    tile_metrics["bytes_read"] = bytes_read
    tile_metrics["write_bps"] = tile_metrics["bytes_read"] / write_time
    tile_metrics["bytes_stored"] = _get_bytes_stored(
        out_zarr, tile_name, n_levels
    )
    storage_ratio = _get_storage_ratio(out_zarr, tile_name)
    tile_metrics["storage_ratio"] = storage_ratio
    LOGGER.info(f"Compress ratio: {storage_ratio}")


def _build_ome(
    data_shape: Tuple[int, ...],
    image_name: str,
    channel_names: Optional[List[str]] = None,
    channel_colors: Optional[List[int]] = None,
    channel_minmax: Optional[List[Tuple[float, float]]] = None,
) -> Dict:
    """
    Create the necessary metadata for an OME tiff image

    Parameters
    ----------
    data_shape: A 5-d tuple, assumed to be TCZYX order
    image_name: The name of the image
    channel_names: The names for each channel
    channel_colors: List of all channel colors
    channel_minmax: List of all (min, max) pairs of channel intensities

    Returns
    -------
    Dict: An "omero" metadata object suitable for writing to ome-zarr
    """
    if channel_names is None:
        channel_names = [
            f"Channel:{image_name}:{i}"
            for i in range(data_shape[1])
        ]
    if channel_colors is None:
        channel_colors = [i for i in range(data_shape[1])]
    if channel_minmax is None:
        channel_minmax = [(0.0, 1.0) for _ in range(data_shape[1])]
    ch = []
    for i in range(data_shape[1]):
        ch.append(
            {
                "active": True,
                "coefficient": 1,
                "color": f"{channel_colors[i]:06x}",
                "family": "linear",
                "inverted": False,
                "label": channel_names[i],
                "window": {
                    "end": float(channel_minmax[i][1]),
                    "max": float(channel_minmax[i][1]),
                    "min": float(channel_minmax[i][0]),
                    "start": float(channel_minmax[i][0]),
                },
            }
        )

    omero = {
        "id": 1,  # ID in OMERO
        "name": image_name,  # Name as shown in the UI
        "version": "0.4",  # Current version
        "channels": ch,
        "rdefs": {
            "defaultT": 0,  # First timepoint to show the user
            "defaultZ": data_shape[2] // 2,  # First Z section to show the user
            "model": "color",  # "color" or "greyscale"
        }
    }
    return omero


def _compute_scales(
    scale_num_levels: int,
    scale_factor: Tuple[float, float, float],
    pixelsizes: Tuple[float, float, float],
    chunks: Tuple[int, int, int, int, int],
    data_shape: Tuple[int, int, int, int, int],
    translation: Optional[List[float]] = None,
) -> Tuple[List, List]:
    """Generate the list of coordinate transformations and associated chunk options.

    Parameters
    ----------
    scale_num_levels: the number of downsampling levels
    scale_factor: a tuple of scale factors in each spatial dimension (Z, Y, X)
    pixelsizes: a list of pixel sizes in each spatial dimension (Z, Y, X)
    chunks: a 5D tuple of integers with size of each chunk dimension (T, C, Z, Y, X)
    data_shape: a 5D tuple of the full resolution image's shape
    translation: a 5 element list specifying the offset in physical units in each dimension

    Returns
    -------
    A tuple of the coordinate transforms and chunk options
    """
    transforms = [
        [
            # the voxel size for the first scale level
            {
                "type": "scale",
                "scale": [
                    1.0,
                    1.0,
                    pixelsizes[0],
                    pixelsizes[1],
                    pixelsizes[2],
                ],
            }
        ]
    ]
    if translation is not None:
        transforms[0].append({
            "type": "translation",
            "translation": translation
        })
    chunk_sizes = []
    lastz = data_shape[2]
    lasty = data_shape[3]
    lastx = data_shape[4]
    opts = dict(
        chunks=(
            1,
            1,
            min(lastz, chunks[2]),
            min(lasty, chunks[3]),
            min(lastx, chunks[4]),
        )
    )
    chunk_sizes.append(opts)
    if scale_num_levels > 1:
        for i in range(scale_num_levels - 1):
            last_transform = transforms[-1][0]
            last_scale = cast(List, last_transform["scale"])
            transforms.append(
                [
                    {
                        "type": "scale",
                        "scale": [
                            1.0,
                            1.0,
                            last_scale[2] * scale_factor[0],
                            last_scale[3] * scale_factor[1],
                            last_scale[4] * scale_factor[2],
                        ],
                    }
                ]
            )
            if translation is not None:
                transforms[-1].append({
                    "type": "translation",
                    "translation": translation
                })
            lastz = int(math.ceil(lastz / scale_factor[0]))
            lasty = int(math.ceil(lasty / scale_factor[1]))
            lastx = int(math.ceil(lastx / scale_factor[2]))
            opts = dict(chunks=(
                1,
                1,
                min(lastz, chunks[2]),
                min(lasty, chunks[3]),
                min(lastx, chunks[4])))
            chunk_sizes.append(opts)

    return transforms, chunk_sizes


def _get_axes_5d(time_unit: str = "millisecond", space_unit: str = "micrometer") -> List[Dict]:
    """Generate the list of axes.

    Parameters
    ----------
    time_unit: the time unit string, e.g., "millisecond"
    space_unit: the space unit string, e.g., "micrometer"

    Returns
    -------
    A list of dictionaries for each axis
    """
    axes_5d = [
        {"name": "t", "type": "time", "unit": f"{time_unit}"},
        {"name": "c", "type": "channel"},
        {"name": "z", "type": "space", "unit": f"{space_unit}"},
        {"name": "y", "type": "space", "unit": f"{space_unit}"},
        {"name": "x", "type": "space", "unit": f"{space_unit}"},
    ]
    return axes_5d


"""
Script to validate smartspim datasets
"""

import logging
import multiprocessing
import os
import re
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List, Union

import exiftool
from argschema import ArgSchema, ArgSchemaParser
from argschema.fields import Boolean, InputDir
from tqdm import tqdm

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s : %(message)s",
    datefmt="%Y-%m-%d %H:%M",
    handlers=[
        logging.StreamHandler(),
        # logging.FileHandler("test.log", "a"),
    ],
)
logging.disable("DEBUG")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

PathLike = Union[Path, str]

POST_PATH_NEW_CONV = "SmartSPIM"
POST_PATH_OLD_CONV = ""


class ValidationParameters(ArgSchema):
    """
    Validation parameters
    """

    path = InputDir(
        required=True,
        metadata={
            "description": "Path where the channels of the dataset are located."
        },
    )

    val_mdata = Boolean(
        required=False,
        metadata={
            "description": """
            Validated the metadata for each of the images.
            Validates file format and bit depth.
            """
        },
        dump_default=False,
    )


class SmartSPIMReader:
    """Reader for smartspim datasets"""

    class RegexPatterns(Enum):
        """Enum for regex patterns for the smartSPIM data"""

        # regex expressions for not structured smartspim datasets
        capture_date_regex = r"(20[0-9]{2}([0-9][0-9]{1})([0-9][0-9]{1}))"
        capture_time_regex = r"(_(\d{2})_(\d{2})_(\d{2})_)"
        capture_mouse_id = r"(_(\d+|[a-zA-Z]*\d+)$)"

        # Regular expression for smartspim datasets
        smartspim_regex = r"SmartSPIM_(\d+|[a-zA-Z]*\d+)_(20\d{2}-(\d\d{1})-(\d\d{1}))_((\d{2})-(\d{2})-(\d{2}))"
        smartspim_old_regex = r"(20\d{2}(\d\d{1})(\d\d{1}))_((\d{2}))_((\d{2}))_((\d{2}))_(\d+|[a-zA-Z]*\d+)"

        # Regex expressions for inner folders inside root
        regex_channels = r"Ex_(\d{3})_Em_(\d{3})$"
        regex_channels_MIP = r"Ex_(\d{3})_Em_(\d{3}_MIP)$"
        regex_files = r'[^"]*.(txt|ini)$'

    @staticmethod
    def read_smartspim_folders(path: PathLike) -> List[str]:
        """
        Reads smartspim datasets in a folder
        based on data conventions

        Parameters
        -----------------
        path: PathLike
            Path where the datasets are located

        Returns
        -----------------
        List[str]
            List with the found smartspim datasets
        """
        smartspim_datasets = []

        if os.path.isdir(path):
            datasets = os.listdir(path)

            for dataset in datasets:
                dataconvention_match = re.match(
                    SmartSPIMReader.RegexPatterns.smartspim_regex.value,
                    dataset,
                )

                oldconvention_match = re.match(
                    SmartSPIMReader.RegexPatterns.smartspim_old_regex.value,
                    dataset,
                )

                if dataconvention_match:
                    str_path = str(Path(dataset).joinpath(POST_PATH_NEW_CONV))
                    smartspim_datasets.append(str_path)

                if oldconvention_match:
                    str_path = str(Path(dataset).joinpath(POST_PATH_OLD_CONV))
                    smartspim_datasets.append(str_path)

        else:
            raise ValueError(f"Path {path} is not a folder.")

        return smartspim_datasets


def save_string_to_txt(txt: str, filepath: PathLike, mode="w") -> None:
    """
    Saves a text in a file in the given mode.
    Parameters
    ------------------------
    txt: str
        String to be saved.
    filepath: PathLike
        Path where the file is located or will be saved.
    mode: str
        File open mode.
    """

    with open(filepath, mode) as file:
        file.write(txt + "\n")


def read_image_directory_structure(folder_dir) -> dict:
    """
    Creates a dictionary representation of all the images
    saved by folder/col_N/row_N/images_N.[file_extention]

    Parameters
    ------------------------
    folder_dir:PathLike
        Path to the folder where the images are stored

    Returns
    ------------------------
    dict:
        Dictionary with the image representation where:
        {channel_1: ... {channel_n: {col_1: ... col_n: {row_1: ... row_n: integer with n images} } } }
    """

    directory_structure = {}
    folder_dir = Path(folder_dir)

    channel_paths = [
        folder_dir.joinpath(folder)
        for folder in os.listdir(folder_dir)
        if os.path.isdir(folder_dir.joinpath(folder))
        and re.match(
            SmartSPIMReader.RegexPatterns.regex_channels.value, folder
        )
    ]

    for channel_idx in range(len(channel_paths)):
        directory_structure[channel_paths[channel_idx]] = {}
        cols = os.listdir(channel_paths[channel_idx])
        # logger.info(f"Validating channel {channel_paths[channel_idx]}")

        for col in tqdm(cols):
            possible_col = channel_paths[channel_idx].joinpath(col)

            if os.path.isdir(possible_col):
                directory_structure[channel_paths[channel_idx]][col] = {}
                rows = os.listdir(possible_col)

                for row in rows:
                    possible_row = (
                        channel_paths[channel_idx].joinpath(col).joinpath(row)
                    )

                    if os.path.isdir(possible_row):
                        col_row_images = os.listdir(possible_row)
                        directory_structure[channel_paths[channel_idx]][col][
                            row
                        ] = col_row_images

    return directory_structure


def get_images_channel(channel_dict: dict) -> int:
    """
    Gets the number of images
    in a channel

    Parameters
    ------------
    channel_dict: dict
        Dictionary with the folder structure

    Returns
    -----------
    Number of images in the channel
    """
    n_images = 0

    for col_name, rows in channel_dict.items():
        for row_name, images in rows.items():
            len_images = len(images)

            if images == len_images:
                raise ValueError(
                    f"Possible error in pos {col_name}/{row_name}"
                )

            n_images += len_images

    return n_images


def get_image_metadata(image_paths: List[PathLike]) -> List[dict]:
    """
    Retrieves image metadata for all the
    images included in the list

    Parameters
    ----------
    image_paths: List[PathLike]
        List with the paths of the images

    Returns
    ----------
    List[dict]
        List with the metadata of each
        image
    """
    with exiftool.ExifToolHelper() as et:
        metadata = et.get_metadata(image_paths)

    return metadata


def validate_rows(
    row_names: List[str],
    row_images: List[str],
    channel_path: PathLike,
    col_name: str,
    file_format: str,
    bit_depth: int,
) -> bool:
    """
    Function that validates the rows
    of a channel in a dataset. Function designed
    to work in parallel
    """

    n_images = len(row_images)

    for n_image in range(n_images):
        print(f"Validating: {col_name}/{row_names[n_image]}")
        image_paths = [
            str(
                Path(channel_path).joinpath(
                    f"{col_name}/{row_names[n_image]}/{image_path}"
                )
            )
            for image_path in row_images[n_image]
        ]

        metadata_files = get_image_metadata(image_paths)
        logger.info(f"Validating metadata: {col_name}/{row_names[n_image]}")

        for metadata_file in metadata_files:
            if metadata_file["File:FileType"] != file_format:
                msg = f"Error in file format for {metadata_file['File:SourceFile']}"
                logger.error(msg)
                return False

            elif metadata_file[f"{file_format}:BitDepth"] != bit_depth:
                msg = f"Error in bit depth for {metadata_file['File:SourceFile']}"
                logger.error(msg)
                return False

    return True


def _validate_rows(args_dict: dict) -> bool:
    """
    Function used to be dispatched to workers
    by using multiprocessing
    """
    return validate_rows(**args_dict)


def validate_metadata_parallel(
    channel_path: str, channel_dict: dict, file_format: str, bit_depth: int
) -> bool:
    """
    Validates image metadata using parallel processing

    Parameters
    ----------
    channel_path: str
        Path where the channel is stored

    channel_dict: dict
        Dicitonary with the channel folder structure
        pointing to row/col/tile.png

    file_format: str
        Expected file format of the images

    bit_depth: int
        Expected bit depth of the images

    Returns
    ----------
    bool
        False if the channel has no issues with tiles,
        True otherwise.
    """
    workers = multiprocessing.cpu_count()
    logger.info(f"N CPUs {workers}")
    rows_per_worker = 1

    for col_name, rows in channel_dict.items():
        n_rows = len(rows)

        if n_rows < workers:
            workers = n_rows
            rows_per_worker = 1
        else:
            if n_rows % 2 != 0:
                rows_per_worker = (n_rows // workers) + 1
            else:
                rows_per_worker = n_rows // workers

        start_row = 0
        end_row = rows_per_worker
        args = []

        row_names = list(rows.keys())
        row_images = list(rows.values())

        for idx_worker in range(workers):
            arg_dict = {
                "row_names": row_names[start_row:end_row],
                "row_images": row_images[start_row:end_row],
                "channel_path": channel_path,
                "col_name": col_name,
                "file_format": file_format,
                "bit_depth": bit_depth,
            }

            args.append(arg_dict)

            if idx_worker + 1 == workers - 1:
                start_row = end_row
                end_row = n_rows
            else:
                start_row = end_row
                end_row += rows_per_worker

        res = []

        with multiprocessing.Pool(workers) as pool:
            results = pool.imap(
                _validate_rows,
                args,
                chunksize=1,
            )

            for pos in results:
                res.append(pos)

        for res_idx in range(len(res)):
            if not res[res_idx]:
                logger.error(
                    f"Dataset with format or bit depth issues found by worker {res_idx}"
                )
                return False

    return True


def validate_metadata(
    channel_path: str, channel_dict: dict, file_format: str, bit_depth: int
) -> bool:
    """
    Validates image metadata

    Parameters
    ----------
    channel_path: str
        Path where the channel is stored

    channel_dict: dict
        Dicitonary with the channel folder structure
        pointing to row/col/tile.png

    file_format: str
        Expected file format of the images

    bit_depth: int
        Expected bit depth of the images

    Returns
    ----------
    bool
        False if the channel has no issues with tiles,
        True otherwise.
    """

    workers = multiprocessing.cpu_count()
    logger.info(f"N CPU cores: {workers}")
    rows_per_worker = 1

    for col_name, rows in channel_dict.items():
        for row_name, images in rows.items():
            print(f"Validating: {col_name}/{row_name}")
            start_date = datetime.now()
            image_paths = [
                str(
                    Path(channel_path).joinpath(
                        f"{col_name}/{row_name}/{image_path}"
                    )
                )
                for image_path in images
            ]

            metadata_files = get_image_metadata(image_paths)

            for metadata_file in metadata_files:
                if metadata_file["File:FileType"] != file_format:
                    msg = f"Error in file format for {metadata_file['File:SourceFile']}"
                    raise ValueError(msg)

                elif metadata_file[f"{file_format}:BitDepth"] != bit_depth:
                    msg = f"Error in bit depth for {metadata_file['File:SourceFile']}"
                    raise ValueError(msg)
            end_date = datetime.now()

            print(f"Time to validate stack of tiles: {end_date - start_date}")

    return True


def validate_dataset(
    dataset_path: PathLike, validate_mdata: bool = False
) -> bool:
    """
    Validates a dataset

    Parameters
    ------------
    dataset_path: PathLike
        Path where the dataset is stored

    Returns
    -----------
    True if the dataset is correct, False otherwise
    """
    dataset_structure = read_image_directory_structure(dataset_path)

    # logger.info(f"Time reading folder structure: {end_time - start_time}")

    images_per_channel = []

    for channel_name, image_paths in dataset_structure.items():
        n_images = get_images_channel(image_paths)
        images_per_channel.append(n_images)
        logger.info(f"Channel {channel_name} has {n_images} images")

    n_channels = len(images_per_channel)

    if not n_channels:
        return False

    channel_images = images_per_channel[0]

    for images_idx in range(1, n_channels):
        if channel_images != images_per_channel[images_idx]:
            return False

    if validate_mdata:
        logger.info("Validating metadata")

        for channel_name, config in dataset_structure.items():
            validation_config = {
                "channel_path": str(channel_name),
                "channel_dict": config,
                "file_format": "PNG",
                "bit_depth": 16,
            }

            channel_val = validate_metadata_parallel(**validation_config)

            if not channel_val:
                logger.error(f"Wrong metadata in channel {channel_name}")
                return False

    return True


def usage_example():
    """
    Usage example function. Please, provide your paths!
    """

    BASE_PATH = Path("BASE_PATH")
    DATASET_PATH = BASE_PATH.joinpath("DATASET_PATH")

    error_dataset_paths = "PATH_TO_OUTPUT_LOG_ERRORS.txt"

    search_datasets = True
    validate_mdata = False
    dataset_paths = []

    if search_datasets:
        dataset_paths = SmartSPIMReader.read_smartspim_folders(BASE_PATH)

    else:
        dataset_paths = [
            # Path where the channels are
            # "SmartSPIM_634571_2022-08-24_19-14-17/SmartSPIM"
            "20230220_12_16_27_642480"
        ]

    datasets_with_problems = ["Datasets with errors:\n"]
    check_paths = ["\nCheck paths:\n"]

    for dataset_path in dataset_paths:
        val_path = Path(BASE_PATH.joinpath(dataset_path))

        logger.info(f"Validating dataset: {dataset_path}")

        try:
            tile_status = validate_dataset(
                val_path, validate_metadata=validate_mdata
            )

        except FileNotFoundError as e:
            logger.error(
                f"[!!] Check path {val_path}. This folder MUST have the channels!"
            )
            check_paths.append(str(val_path))

        if not tile_status:
            logger.error(f"\n[+] Dataset {dataset_path} has problems in tiles")
            datasets_with_problems.append(str(dataset_path))

        else:
            logger.info(f"\n[+] Dataset {dataset_path} does not have issues")

    # Saving datasets with errors
    join_lists = datasets_with_problems + check_paths
    txt = "\n".join(join_lists)
    save_string_to_txt(txt, error_dataset_paths)


def main():
    """
    Main function to execute individual dataset validation
    """
    # usage_example()
    mod = ArgSchemaParser(schema_type=ValidationParameters)
    args = mod.args

    # Unpacking parameters
    path = args["path"]
    validate_mdata = args["val_mdata"]

    logger.info(f"Validating {path}")

    if validate_dataset(dataset_path=path, validate_mdata=validate_mdata):
        logger.info(f"Dataset {path} is valid.")

    else:
        logger.error(f"Dataset {path} has problems with tiles.")


if __name__ == "__main__":
    main()



import logging
import xml.etree.ElementTree as ET
from pathlib import Path

import numpy as np
from packaging.version import parse

# geometry properties for NP-opto
X_PITCH = 48
X_OFFSET = 11
Y_PITCH = 20
NUMEL_IN_COL = 192


def get_standard_np_opto_electrode_positions():
    npopto_electrode_xpos_arr = [X_OFFSET, X_OFFSET + X_PITCH] * NUMEL_IN_COL
    npopto_electrode_ypos_arr = np.concatenate(
        [i * np.array([Y_PITCH, Y_PITCH]) for i in range(NUMEL_IN_COL)]
    )
    npopto_electrode_xpos = {}
    npopto_electrode_ypos = {}
    for ch, (xpos, ypos) in enumerate(
        zip(npopto_electrode_xpos_arr, npopto_electrode_ypos_arr)
    ):
        npopto_electrode_xpos[f"CH{ch}"] = str(xpos)
        npopto_electrode_ypos[f"CH{ch}"] = str(ypos)
    return npopto_electrode_xpos, npopto_electrode_ypos


def correct_np_opto_electrode_locations(input_dir):
    # find settings files
    input_dir = Path(input_dir)
    settings_files = list(input_dir.glob("**/*.xml"))

    (
        npopto_electrode_xpos,
        npopto_electrode_ypos,
    ) = get_standard_np_opto_electrode_positions()

    for settings_file in settings_files:
        # parse xml
        tree = ET.parse(str(settings_file))
        root = tree.getroot()

        signal_chain = root.find("SIGNALCHAIN")
        neuropix_pxi = None
        for processor in signal_chain:
            if "PROCESSOR" == processor.tag:
                name = processor.attrib["name"]
                if "Neuropix-PXI" in name:
                    neuropix_pxi = processor
                    break

        if neuropix_pxi is None:
            return
        neuropix_pxi_version = parse(neuropix_pxi.attrib["libraryVersion"])
        if neuropix_pxi_version > parse("0.4.0"):
            # no correction needed for this plugin version
            return

        editor = neuropix_pxi.find("EDITOR")
        np_probes = editor.findall("NP_PROBE")

        needs_correction = False
        for np_probe in np_probes:
            if "OPTO" in np_probe.attrib["headstage_part_number"]:
                logging.info("Found NP-OPTO!")
                needs_correction = True

                # update channel locations
                electrode_xpos = np_probe.find("ELECTRODE_XPOS")
                electrode_ypos = np_probe.find("ELECTRODE_YPOS")
                electrode_xpos.attrib = npopto_electrode_xpos
                electrode_ypos.attrib = npopto_electrode_ypos

        settings_file_path = str(settings_file)
        if needs_correction:
            wrong_settings_file = settings_file.parent / "settings.xml.wrong"
            logging.info(
                f"Renaming wrong NP-OPTO settings file as "
                f"{wrong_settings_file}"
            )
            settings_file.rename(wrong_settings_file)
            logging.info(
                f"Saving correct NP-OPTO settings file as "
                f"{settings_file_path}"
            )
            tree.write(settings_file_path)
        else:
            logging.info(f"No NP-OPTO probes found in {settings_file_path}")



import fnmatch
import json
import os
import re
import shutil
import subprocess
from pathlib import Path, PurePath, PurePosixPath
from typing import Any, List, Optional, Tuple, Union

import yaml

from aind_data_transfer.util.io_utils import DataReaderFactory

PathLike = Union[str, Path]


def collect_filepaths(
    folder: Union[str, os.PathLike],
    recursive: bool = True,
    include_exts: Optional[List[str]] = None,
    exclude_dirs: Optional[List[str]] = None,
) -> List[str]:
    """Get the absolute paths for all files in folder
    Args:
        folder (str): the directory to look for files
        recursive (bool): whether to traverse all sub-folders
        include_exts (optional): list of valid file extensions to include.
                                 e.g., ['.tiff', '.h5', '.ims']
        exclude_dirs (optional): list of directories to exclude from the search
    Returns:
        list of filepaths
    """
    if exclude_dirs is None:
        exclude_dirs = []
    filepaths = []
    for root, _, files in os.walk(folder):
        root_name = Path(root).name
        if root_name in exclude_dirs:
            continue
        for f in files:
            path = os.path.join(root, f)
            _, ext = os.path.splitext(path)
            if include_exts is None or ext in include_exts:
                filepaths.append(path)
        if not recursive:
            break
    return filepaths


def get_images(
    image_folder: Union[str, os.PathLike],
    exclude: List[str] = None,
    include_exts: List[str] = DataReaderFactory().VALID_EXTENSIONS,
    recursive: bool = False,
) -> List[str]:
    """Get the absolute paths for all images in a folder
    Args:
        image_folder: the directory to look for images
        exclude: list of filename patterns to exclude
        include_exts: list of valid file extensions to include.
                                 e.g., ['.tiff', '.h5', '.ims']
        recursive: whether to traverse all sub-folders
    Returns:
        list of image paths
    """
    if exclude is None:
        exclude = []
    image_paths = collect_filepaths(
        image_folder,
        recursive=recursive,
        include_exts=include_exts,
    )

    exclude_paths = set()
    for path in image_paths:
        if any(fnmatch.fnmatch(path, pattern) for pattern in exclude):
            exclude_paths.add(path)

    image_paths = [p for p in image_paths if p not in exclude_paths]

    return image_paths


def join_cloud_paths(cloud_dest_path: str, relpath: str) -> str:
    """Always produce posix-style paths, even if relpath
    is Windows-style
    Args:
        cloud_dest_path (str): first part of the path
        relpath (str): second part of the path
    Returns:
        the joined path
    """
    cloud_dest_path = PurePosixPath(cloud_dest_path)
    relpath = PurePath(relpath)
    return str(cloud_dest_path / relpath)


def make_cloud_paths(
    filepaths: List[Union[str, os.PathLike]],
    cloud_dest_path: Union[str, os.PathLike],
    root: Union[str, os.PathLike] = None,
) -> List[str]:
    """
    Given a list of filepaths and a cloud destination folder,
    build a cloud path for each file relative to root.
    Args:
        filepaths (list): list of paths
        cloud_dest_path (str): the cloud storage path to store files
        root (str): a directory shared by all paths in filepaths, which will
                    serve as the new root under cloud_dest path. If none,
                    all files are uploaded as a flat list to cloud_dest_path,
                    ignoring any exiting directory structure.
    Returns:
        list of cloud storage paths
    Examples:
    >>> filepaths = ["/data/micr/0001.tif", "/data/metadata/rig.json"]
    >>> root = "/data"
    >>> cloud_dest_path = "my-data"
    >>> cloud_paths = make_cloud_paths(filepaths, cloud_dest_path, root)
    >>> print(cloud_paths)
    ['my-data/micr/001.tif', 'my-data/metadata/rig.json']
    >>> cloud_paths = make_cloud_paths(filepaths, cloud_dest_path, None)
    >>> print(cloud_paths)
    ['my-data/001.tif', 'my-data/rig.json']
    """
    cloud_paths = []
    # remove both leading and trailing '/'
    cloud_dest_path = cloud_dest_path.strip("/")
    for fpath in filepaths:
        if root is None:
            cloud_paths.append(
                join_cloud_paths(cloud_dest_path, PurePath(fpath).name)
            )
        else:
            cloud_paths.append(
                join_cloud_paths(cloud_dest_path, os.path.relpath(fpath, root))
            )
    return cloud_paths


def is_cloud_url(url: str):
    """
    Test if the url points to an AWS S3 or Google Cloud Storage URI
    Args:
        url: the url to test
    Returns:
        True if url is a cloud url
    """
    url = str(url)
    if url.startswith("s3://"):
        return True
    if url.startswith("gs://"):
        return True
    return False


def parse_cloud_url(cloud_url: str) -> Tuple[str, str, str]:
    """
    Get the cloud storage provider, bucket name, and path
    from an AWS S3 or Google Cloud Storage url.
    Args:
        cloud_url: the cloud url to parse
    Returns:
        a tuple containing the provider, bucket and path
    """
    parts = Path(cloud_url).parts
    provider = parts[0] + "//"
    bucket = parts[1]
    cloud_dst = "/".join(parts[2:])
    return provider, bucket, cloud_dst


def create_folder(dest_dir: PathLike, verbose: Optional[bool] = False) -> None:
    """
    Create new folders.

    Parameters
    ------------------------
    dest_dir: PathLike
        Path where the folder will be created if it does not exist.
    verbose: Optional[bool]
        If we want to show information about the folder status. Default False.

    Raises
    ------------------------
    OSError:
        if the folder exists.

    """

    if not (os.path.exists(dest_dir)):
        try:
            if verbose:
                print(f"Creating new directory: {dest_dir}")
            os.makedirs(dest_dir, mode=0o777)
        except OSError as e:
            raise


def delete_folder(dest_dir: PathLike) -> None:
    """
    Deletes a folder.

    Parameters
    ------------------------
    dest_dir: PathLike
        Path where the folder will be created if it does not exist.

    Raises
    ------------------------
    OSError:
        if the folder could not be deleted.
    """

    if os.path.exists(dest_dir):
        try:
            shutil.rmtree(dest_dir)
            print(f"Folder {dest_dir} was removed!")
        except shutil.Error as e:
            print(f"Folder could not be removed! Error {e}")


def write_list_to_txt(filename_path: PathLike, list_values: List) -> None:
    """
    Writes a list to a text file.

    Parameters
    ------------------------
    filename_path: PathLike
        Path where the file will be created.

    list_values: List
        List with the values to be written.
    """

    with open(filename_path, "w") as dicti_file:
        for value in list_values:
            dicti_file.write(f"{value}\n")


def move_folders_or_files(
    orig_path: PathLike,
    dest_path: PathLike,
    regex_folders: str,
    mode: Optional[str] = "move",
) -> None:
    """
    Move/copy folders or files to another location.

    Parameters
    ------------------------
    orig_path: PathLike
        Path where the folders/files are located.

    dest_path: PathLike
        Path where the folders/files will be located.

    regex_folders:str
        Regular expression to get folders/files

    mode:str
        Mode to move data. It could be move or copy.

    """

    orig_path = Path(orig_path)
    dest_path = Path(dest_path)

    # Convert to regular expression format
    regex_folders = "({})".format(regex_folders)

    if not os.path.isdir(orig_path) or not os.path.isdir(dest_path):
        raise ValueError("Please, check source and destination paths.")

    else:
        elements = [
            element
            for element in os.listdir(orig_path)
            if re.match(regex_folders, element)
        ]

        for element in elements:
            move_path = orig_path.joinpath(element)

            if mode == "move":
                shutil.move(str(move_path), str(dest_path))
            elif mode == "copy":
                dest_copy_path = dest_path.joinpath(element)

                if os.path.isdir(move_path):
                    shutil.copytree(str(move_path), str(dest_copy_path))

                elif os.path.isfile(move_path):
                    shutil.copyfile(str(move_path), str(dest_copy_path))

                else:
                    raise ValueError(
                        f"Element {element} is not a file nor a directory."
                    )

            else:
                raise NotImplementedError(
                    f"Mode {mode} has not been implemented."
                )


def check_path_instance(obj: object) -> bool:
    """
    Checks if an objects belongs to pathlib.Path subclasses.

    Parameters
    ------------------------
    obj: object
        Object that wants to be validated.

    Returns
    ------------------------
    bool:
        True if the object is an instance of Path subclass, False otherwise.
    """

    for childclass in Path.__subclasses__():
        if isinstance(obj, childclass):
            return True

    return False


def save_dict_as_json(
    filename: str, dictionary: dict, verbose: Optional[bool] = False
) -> None:
    """
    Saves a dictionary as a json file.

    Parameters
    ------------------------
    filename: str
        Name of the json file.
    dictionary: dict
        Dictionary that will be saved as json.
    verbose: Optional[bool]
        True if you want to print the path where the file was saved.

    """

    if dictionary == None:
        dictionary = {}

    else:
        for key, value in dictionary.items():
            # Converting path to str to dump dictionary into json
            if check_path_instance(value):
                # TODO fix the \\ encode problem in dump
                dictionary[key] = str(value)

    with open(filename, "w") as json_file:
        json.dump(dictionary, json_file, indent=4)

    if verbose:
        print(f"- Json file saved: {filename}")


def execute_command(command: str, print_command: bool = False) -> None:
    """
    Execute a shell command.

    Parameters
    ------------------------
    command: str
        Command that we want to execute.
    print_command: bool
        Bool that dictates if we print the command in the console.

    Raises
    ------------------------
    CalledProcessError:
        if the command could not be executed (Returned non-zero status).

    """

    if print_command:
        print(command)

    popen = subprocess.Popen(
        command, stdout=subprocess.PIPE, universal_newlines=True, shell=True
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        yield str(stdout_line).strip()
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)


def read_text_to_list(path: str):
    """
    Function to get the information saved in a file.
    Parameters:
      - path: Path where the file is located.
    Returns:
      - list
    """

    if not os.path.isfile(path):
        return False

    list_text = []

    with open(path, "r") as info_file:
        while True:
            text = info_file.readline().rstrip("\r\n")

            if not text:
                break
            else:
                list_text.append(text)

    return list_text


def get_status_filename_data(dataset_path: PathLike) -> list:
    """
    Checks the status filename data.

    Parameters
    ------------------------
    dataset_path: PathLike
        Path where the dataset will be located.

    Returns
    ------------------------
    List:
        Text file content.
    """
    STATUS_FILENAME = "DATASET_STATUS.txt"
    file_content = []

    if os.path.isdir(dataset_path):
        filename_path = [
            dataset_path.joinpath(f)
            for f in os.listdir(dataset_path)
            if f == STATUS_FILENAME
            and os.path.isfile(os.path.join(dataset_path, f))
        ]
        if not len(filename_path):
            return []

        file_content = read_text_to_list(filename_path[0])

    return file_content


def read_json_as_dict(filepath: str) -> dict:
    """
    Reads a json as dictionary.
    Parameters
    ------------------------
    filepath: PathLike
        Path where the json is located.
    Returns
    ------------------------
    dict:
        Dictionary with the data the json has.
    """

    dictionary = {}

    if os.path.exists(filepath):
        with open(filepath) as json_file:
            dictionary = json.load(json_file)

    return dictionary


def save_dict_as_json(filename: str, dictionary: dict) -> None:
    """
    Saves a dictionary as a json file.
    Parameters
    ------------------------
    filename: str
        Name of the json file.
    dictionary: dict
        Dictionary that will be saved as json.
    """

    if dictionary is None:
        dictionary = {}

    with open(filename, "w") as json_file:
        json.dump(dictionary, json_file, indent=4)


def update_json_key(json_path: PathLike, key: Any, new_value: Any):
    """
    Updates a key of a json

    Parameters
    ----------
    json_path: PathLike
        Path where the json is located

    key: Any
        Attribute to modify

    new_value: Any
        New value to add

    Raises
    ----------
    FileNotFoundError
        If the file does not exist, a file not found
        error is thrown.
    """

    json_path = str(json_path)
    if not os.path.exists(json_path):
        raise FileNotFoundError(f"{json_path} does not exist!")

    json_data = read_json_as_dict(json_path)

    json_data[key] = new_value
    save_dict_as_json(json_path, json_data)


def write_dict_to_yaml(dictionary: dict, filename: PathLike) -> None:
    """
    Writes a dictionary to a YAML file.

    Parameters
    ------------------------
    dictionary: dict
        Dictionary with the data to be written.

    filename: PathLike
        Path where the YAML will be saved
    """

    with open(filename, "w") as outfile:
        yaml.dump(dictionary, outfile, default_flow_style=False)


def helper_validate_key_dict(
    dictionary: dict, key: Any, default_return: Any = None
) -> Any:
    """
    Helper function that validates if a key is
    in a dictionary
    Parameters
    ----------
    dictionary: dict
        Dictionary from which we want to extract
        the value of a key
    key: Any
        Key of the dictionary

    default_return: Any
        Default return of the function
        if the key does not exist.
        Default None

    Returns
    ---------
    Any
        Value of the key in the
        dictionary
    """
    if key in dictionary:
        return dictionary[key]

    return default_return


def any_hdf5(filepaths: List[str]) -> bool:
    """
    Checks if any filepaths are HDF5-based, e.g., ".h5", ".ims"

    Args:
        filepaths: a list of filepath strings

    Returns:
        True if any files are HDF5-based, else False
    """
    return any(fp.endswith((".h5", ".ims")) for fp in filepaths)



import unittest

import dask.array as da
import numpy as np

from aind_data_transfer.transformations.deinterleave import Deinterleave


def _create_interleaved_array(shape, num_channels, array_type):
    if array_type == "numpy":
        a = np.zeros(shape, dtype=int)
    elif array_type == "dask":
        a = da.zeros(shape, dtype=int)
    else:
        raise ValueError(f"Invalid type {array_type}")
    for offset in range(num_channels):
        a[offset::num_channels, ...] = offset
    return a


class TestDeinterleave(unittest.TestCase):

    def test_deinterleave_dask(self):
        num_channels = 3
        a = _create_interleaved_array((384, 32, 32), num_channels, "dask")
        channels = Deinterleave.deinterleave(a, num_channels, axis=0)
        self.assertEqual(len(channels), 3)
        expected_shape = (128, 32, 32)
        for i, c in enumerate(channels):
            self.assertEqual(c.shape, expected_shape)
            self.assertEqual(c.sum().compute(), np.product(c.shape) * i)

    def test_deinterleave_numpy(self):
        num_channels = 3
        a = _create_interleaved_array((384, 32, 32), num_channels, "numpy")
        channels = Deinterleave.deinterleave(a, num_channels, axis=0)
        self.assertEqual(len(channels), 3)
        expected_shape = (128, 32, 32)
        for i, c in enumerate(channels):
            self.assertEqual(c.shape, expected_shape)
            self.assertEqual(c.sum(), np.product(c.shape) * i)

    def test_bad_shape(self):
        num_channels = 2
        a = _create_interleaved_array((257, 32, 32), num_channels, "numpy")
        channels = Deinterleave.deinterleave(a, num_channels, axis=0)
        self.assertEqual(len(channels), 2)
        expected_shape_ch0 = (129, 32, 32)
        expected_shape_ch1 = (128, 32, 32)
        self.assertEqual(channels[0].shape, expected_shape_ch0)
        self.assertEqual(channels[1].shape, expected_shape_ch1)



"""This module adds classes to handle resolving common endpoints used in the
data transfer jobs."""

import argparse
import json
import os
import re
from datetime import date, datetime, time
from pathlib import Path
from typing import Any, Dict, List, Optional

from aind_data_access_api.secrets import get_parameter
from aind_data_schema.data_description import (
    ExperimentType,
    Modality,
    build_data_name,
)
from aind_data_schema.processing import ProcessName
from pydantic import (
    BaseSettings,
    DirectoryPath,
    Field,
    FilePath,
    PrivateAttr,
    SecretStr,
    validator,
)


class BasicJobEndpoints(BaseSettings):
    """Endpoints that define the services to read/write from"""

    aws_param_store_name: Optional[str] = Field(default=None, repr=False)

    codeocean_domain: str = Field(...)
    codeocean_trigger_capsule_id: str = Field(...)
    codeocean_trigger_capsule_version: Optional[str] = Field(None)
    metadata_service_domain: str = Field(...)
    aind_data_transfer_repo_location: str = Field(...)
    video_encryption_password: Optional[SecretStr] = Field(None)
    codeocean_api_token: Optional[SecretStr] = Field(None)
    codeocean_process_capsule_id: Optional[str] = Field(
        None,
        description=(
            "If defined, will run this Code Ocean Capsule after registering "
            "the data asset"
        ),
    )

    class Config:
        """This class will add custom sourcing from aws."""

        @staticmethod
        def settings_from_aws(param_name: Optional[str]):  # noqa: C901
            """
            Curried function that returns a function to retrieve creds from aws
            Parameters
            ----------
            param_name : Name of the credentials we wish to retrieve
            Returns
            -------
            A function that retrieves the credentials.
            """

            def set_settings(_: BaseSettings) -> Dict[str, Any]:
                """
                A simple settings source that loads from aws secrets manager
                """
                params_from_aws = json.loads(get_parameter(param_name))
                if params_from_aws.get("video_encryption_password_path"):
                    video_encrypt_pwd = json.loads(
                        get_parameter(
                            params_from_aws.get(
                                "video_encryption_password_path"
                            ),
                            with_decryption=True,
                        )
                    )
                    params_from_aws[
                        "video_encryption_password"
                    ] = video_encrypt_pwd.get("password")
                    if params_from_aws.get("video_encryption_password_path"):
                        del params_from_aws["video_encryption_password_path"]
                if params_from_aws.get("codeocean_api_token_path"):
                    co_api_token = json.loads(
                        get_parameter(
                            params_from_aws.get("codeocean_api_token_path"),
                            with_decryption=True,
                        )
                    )
                    params_from_aws["codeocean_api_token"] = co_api_token.get(
                        "CODEOCEAN_READWRITE_TOKEN"
                    )
                    if params_from_aws.get("codeocean_api_token_path"):
                        del params_from_aws["codeocean_api_token_path"]
                return params_from_aws

            return set_settings

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            """Class method to return custom sources."""
            aws_param_store_name = init_settings.init_kwargs.get(
                "aws_param_store_name"
            )
            if aws_param_store_name:
                return (
                    init_settings,
                    env_settings,
                    file_secret_settings,
                    cls.settings_from_aws(param_name=aws_param_store_name),
                )
            else:
                return (
                    init_settings,
                    env_settings,
                    file_secret_settings,
                )


class ModalityConfigs(BaseSettings):
    """Class to contain configs for each modality type"""

    # Optional number id to assign to modality config
    _number_id: Optional[int] = PrivateAttr(default=None)
    modality: Modality = Field(
        ..., description="Data collection modality", title="Modality"
    )
    source: DirectoryPath = Field(
        ...,
        description="Location of raw data to be uploaded",
        title="Data Source",
    )
    compress_raw_data: Optional[bool] = Field(
        default=None,
        description="Run compression on data",
        title="Compress Raw Data",
    )
    extra_configs: Optional[FilePath] = Field(
        default=None,
        description="Location of additional configuration file",
        title="Extra Configs",
    )

    @property
    def number_id(self):
        """Retrieve an optionally assigned numerical id"""
        return self._number_id

    @property
    def default_output_folder_name(self):
        """Construct the default folder name for the modality."""
        if self._number_id is None:
            return self.modality.name.lower()
        else:
            return self.modality.name.lower() + str(self._number_id)

    @validator("compress_raw_data", always=True)
    def get_compress_source_default(
        cls, compress_source: Optional[bool], values: Dict[str, Any]
    ) -> bool:
        """Set compress source default to True for ecephys data."""
        if (
            compress_source is None
            and "modality" in values
            and values["modality"] == Modality.ECEPHYS
        ):
            return True
        else:
            return False


class BasicUploadJobConfigs(BasicJobEndpoints):
    """Configuration for the basic upload job"""

    s3_bucket: str = Field(
        ...,
        description="Bucket where data will be uploaded",
        title="S3 Bucket",
    )
    experiment_type: ExperimentType = Field(
        ..., description="Experiment type", title="Experiment Type"
    )
    modalities: List[ModalityConfigs] = Field(
        ...,
        description="Data collection modalities and their directory location",
        title="Modalities",
    )
    subject_id: str = Field(..., description="Subject ID", title="Subject ID")
    acq_date: date = Field(
        ..., description="Date data was acquired", title="Acquisition Date"
    )
    acq_time: time = Field(
        ...,
        description="Time of day data was acquired",
        title="Acquisition Time",
    )
    process_name: ProcessName = Field(
        default=ProcessName.OTHER,
        description="Type of processing performed on the raw data source.",
        title="Process Name",
    )
    temp_directory: Optional[DirectoryPath] = Field(
        default=None,
        description=(
            "As default, the file systems temporary directory will be used as "
            "an intermediate location to store the compressed data before "
            "being uploaded to s3"
        ),
        title="Temp directory",
    )
    behavior_dir: Optional[DirectoryPath] = Field(
        default=None,
        description="Directory of behavior data",
        title="Behavior Directory",
    )
    metadata_dir: Optional[DirectoryPath] = Field(
        default=None,
        description="Directory of metadata",
        title="Metadata Directory",
    )
    log_level: str = Field(
        default="WARNING",
        description="Logging level. Default is WARNING.",
        title="Log Level",
    )
    metadata_dir_force: bool = Field(
        default=False,
        description=(
            "Whether to override metadata from service with metadata in "
            "optional metadata directory"
        ),
        title="Metadata Directory Force",
    )
    dry_run: bool = Field(
        default=False,
        description="Perform a dry-run of data upload",
        title="Dry Run",
    )
    force_cloud_sync: bool = Field(
        default=False,
        description=(
            "Force syncing of data folder even if location exists in cloud"
        ),
        title="Force Cloud Sync",
    )

    @property
    def s3_prefix(self):
        """Construct s3_prefix from configs."""
        return build_data_name(
            label=f"{self.experiment_type.value}_{self.subject_id}",
            creation_date=self.acq_date,
            creation_time=self.acq_time,
        )

    @staticmethod
    def parse_date(date_str: str) -> date:
        """Parses date string to %YYYY-%MM-%DD format"""
        pattern = r"^\d{4}-\d{2}-\d{2}$"
        pattern2 = r"^\d{1,2}/\d{1,2}/\d{4}$"
        if re.match(pattern, date_str):
            return date.fromisoformat(date_str)
        elif re.match(pattern2, date_str):
            return datetime.strptime(date_str, "%m/%d/%Y").date()
        else:
            raise ValueError(
                "Incorrect date format, should be YYYY-MM-DD or MM/DD/YYYY"
            )

    @staticmethod
    def parse_time(time_str: str) -> time:
        """Parses time string to "%HH-%MM-%SS format"""
        pattern = r"^\d{1,2}-\d{1,2}-\d{1,2}$"
        pattern2 = r"^\d{1,2}:\d{1,2}:\d{1,2}$"
        if re.match(pattern, time_str):
            return datetime.strptime(time_str, "%H-%M-%S").time()
        elif re.match(pattern2, time_str):
            return time.fromisoformat(time_str)
        else:
            raise ValueError(
                "Incorrect time format, should be HH-MM-SS or HH:MM:SS"
            )

    @classmethod
    def from_args(cls, args: list):
        """Adds ability to construct settings from a list of arguments."""

        def _help_message(key: str) -> str:
            """Construct help message from field description"""
            return BasicUploadJobConfigs.schema()["properties"][key][
                "description"
            ]

        parser = argparse.ArgumentParser()
        # Required
        parser.add_argument(
            "-a",
            "--acq-date",
            required=True,
            type=str,
            help="Date data was acquired, yyyy-MM-dd or dd/MM/yyyy",
        )
        parser.add_argument(
            "-b",
            "--s3-bucket",
            required=True,
            type=str,
            help=_help_message("s3_bucket"),
        )
        parser.add_argument(
            "-e",
            "--experiment-type",
            required=True,
            type=str,
            help=_help_message("experiment_type"),
        )
        parser.add_argument(
            "-m",
            "--modalities",
            required=True,
            type=str,
            help=(
                f"String that can be parsed as json list where each entry "
                f"has fields: {ModalityConfigs.__fields__}"
            ),
        )
        parser.add_argument(
            "-p",
            "--endpoints-parameters",
            required=True,
            type=str,
            help=(
                "Either a string that can be parsed as a json object or a name"
                " that points to an aws parameter store location"
            ),
        )
        parser.add_argument(
            "-s",
            "--subject-id",
            required=True,
            type=str,
            help=_help_message("subject_id"),
        )
        parser.add_argument(
            "-t",
            "--acq-time",
            required=True,
            type=str,
            help="Time data was acquired, HH-mm-ss or HH:mm:ss",
        )
        # Optional
        parser.add_argument(
            "-l",
            "--log-level",
            required=False,
            type=str,
            help=_help_message("log_level"),
        )
        parser.add_argument(
            "-n",
            "--temp-directory",
            required=False,
            type=str,
            help=_help_message("temp_directory"),
        )
        parser.add_argument(
            "-v",
            "--behavior-dir",
            required=False,
            type=str,
            help=_help_message("behavior_dir"),
        )
        parser.add_argument(
            "-x",
            "--metadata-dir",
            required=False,
            type=str,
            help=_help_message("metadata_dir"),
        )
        parser.add_argument(
            "--dry-run", action="store_true", help=_help_message("dry_run")
        )
        parser.add_argument(
            "--metadata-dir-force",
            action="store_true",
            help=_help_message("metadata_dir_force"),
        )
        parser.add_argument(
            "--force-cloud-sync",
            action="store_true",
            help=_help_message("force_cloud_sync"),
        )
        parser.add_argument(
            "-i",
            "--codeocean-process-capsule-id",
            required=False,
            type=str,
            help=_help_message("codeocean_process_capsule_id"),
        )
        parser.set_defaults(dry_run=False)
        parser.set_defaults(metadata_dir_force=False)
        parser.set_defaults(force_cloud_sync=False)
        job_args = parser.parse_args(args)
        acq_date = BasicUploadJobConfigs.parse_date(job_args.acq_date)
        acq_time = BasicUploadJobConfigs.parse_time(job_args.acq_time)
        behavior_dir = (
            None
            if job_args.behavior_dir is None
            else Path(os.path.abspath(job_args.behavior_dir))
        )
        metadata_dir = (
            None
            if job_args.metadata_dir is None
            else Path(os.path.abspath(job_args.metadata_dir))
        )
        temp_directory = (
            None
            if job_args.temp_directory is None
            else Path(os.path.abspath(job_args.temp_directory))
        )
        log_level = (
            BasicUploadJobConfigs.__fields__["log_level"].default
            if job_args.log_level is None
            else job_args.log_level
        )
        # The user can define the endpoints explicitly as an object that can be
        # parsed with json.loads()
        try:
            params_from_json = BasicJobEndpoints.parse_obj(
                json.loads(job_args.endpoints_parameters)
            )
            endpoints_param_dict = params_from_json.dict()
        # If the endpoints are not defined explicitly, then we can check if
        # the input defines an aws parameter store name
        except json.decoder.JSONDecodeError:
            endpoints_param_dict = {
                "aws_param_store_name": job_args.endpoints_parameters
            }
        if job_args.codeocean_process_capsule_id is not None:
            endpoints_param_dict[
                "codeocean_process_capsule_id"
            ] = job_args.codeocean_process_capsule_id
        modalities_json = json.loads(job_args.modalities)
        modalities = [ModalityConfigs.parse_obj(m) for m in modalities_json]
        return cls(
            s3_bucket=job_args.s3_bucket,
            subject_id=job_args.subject_id,
            experiment_type=ExperimentType(job_args.experiment_type),
            modalities=modalities,
            acq_date=acq_date,
            acq_time=acq_time,
            behavior_dir=behavior_dir,
            temp_directory=temp_directory,
            metadata_dir=metadata_dir,
            dry_run=job_args.dry_run,
            metadata_dir_force=job_args.metadata_dir_force,
            force_cloud_sync=job_args.force_cloud_sync,
            log_level=log_level,
            **endpoints_param_dict,
        )
